#include "clframework.h"

using namespace clframework;

CLGlobalBuffer::CLGlobalBuffer(CLLink* link, std::string name, CLVariableDataType varDataType, unsigned int elements, bool isDynamic, bool paralelize):
  CLVariable(name, GlobalBuffer, varDataType, elements, paralelize),
  link(link),
  isDynamic(isDynamic),
  bufferCount(0),
  clBuffers(NULL),
  partElementCount(NULL),
  offsets(NULL),
  hostHasData(false),
  hostDataChanged(false) {

  if (link == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot create Global buffer for NULL link");
  }

  needsUpdate = true;
}

CLGlobalBuffer::~CLGlobalBuffer() {
  release();
}

bool CLGlobalBuffer::download(bool waitToFinish, bool forceDownload) {
  Log* log = Log::getInstance();

  if (needsUpdate) {
    if (!allocate()) {
      return false;
    }
  }

  log->writeLine(CLDebug, "Reading global buffer: " + name);

  if (data == NULL) {
    if (!allocateHostData()) {
      log->writeLine(CLError, "Cannot allocate data on host");
      return false;
    }
  }

  if (hostHasData && !forceDownload) {
    return true;
  }

  if (memorySize == 0 || link == NULL || clBuffers == NULL) {
    log->writeLine(CLError, "Cannot read uninitialized OpenCL buffer");
    return false;
  }

  cl_int status;
  cl_event* events = NULL;

  if (CLSystem::getInstance()->getProfiling() || waitToFinish) {
    events = new cl_event[bufferCount];
  }

  for (unsigned int i = 0; i < bufferCount; i++) {
    if (events != NULL) {
      status = clEnqueueReadBuffer(link->getQueue(i), clBuffers[i], CL_FALSE, 0, memorySize, data, 0, NULL, &events[i]);
    }
    else {
      status = clEnqueueReadBuffer(link->getQueue(i), clBuffers[i], CL_FALSE, 0, memorySize, data, 0, NULL, NULL);
    }

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while enqueing read buffer: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  if (CLSystem::getInstance()->getProfiling()) {
    clWaitForEvents(bufferCount, events);

    std::ostringstream ss;

    for (unsigned int i = 0; i < bufferCount; i++) {
      cl_ulong timeStart, timeEnd;

      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &timeStart, NULL);
      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &timeEnd, NULL);
      
      ss.str("");
      ss << ((int)(timeEnd - timeStart) / 1000);

      log->writeLine(CLDebug, ss.str() + " ms");
      
      clReleaseEvent(events[i]);
    }

    delete[] events;
  }
  else if (waitToFinish) {
    status = clWaitForEvents(bufferCount, events);

    for (unsigned int i = 0; i < bufferCount; i++) {
      clReleaseEvent(events[i]);
    }

    delete[] events;

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while waiting for events: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  hostHasData = true;
  hostDataChanged = false;

  return true;
}

bool CLGlobalBuffer::upload(bool waitToFinish) {
  Log* log = Log::getInstance();

  if (needsUpdate) {
    if (!allocate()) {
      return false;
    }
  }

  if (data == NULL || !hostHasData) {
    log->writeLine(CLDebug, "!HostHasData");
    return true;
  }

  if (!hostDataChanged) {
    hostHasData = false;
    return true;
  }

  log->writeLine(CLDebug, "Writing global buffer: " + name);

  if (memorySize == 0 || link == NULL || clBuffers == NULL) {
    log->writeLine(CLError, "Cannot write to uninitialized OpenCL buffer");
    return false;
  }

  cl_int status;
  cl_event* events = NULL;

  if (CLSystem::getInstance()->getProfiling() || waitToFinish) {
    events = new cl_event[bufferCount];
  }

  size_t typeSize = CLSystem::dataTypeSize(varDataType);

  for (unsigned int i = 0; i < bufferCount; i++) {
    size_t offsetBytes = getOffset(i) * typeSize;
    char *offsetPos = data + offsetBytes;

    if (events != NULL) {
      status = clEnqueueWriteBuffer(link->getQueue(i), clBuffers[i], CL_FALSE, offsetBytes, getElementCount(i)*typeSize, offsetPos, 0, NULL, &events[i]);
    }
    else {
      status = clEnqueueWriteBuffer(link->getQueue(i), clBuffers[i], CL_FALSE, offsetBytes, getElementCount(i)*typeSize, offsetPos, 0, NULL, NULL);
    }

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while enqueing write buffer: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  if (CLSystem::getInstance()->getProfiling()) {
    clWaitForEvents(bufferCount, events);

    std::ostringstream ss;

    for (unsigned int i = 0; i < bufferCount; i++) {
      cl_ulong timeStart, timeEnd;

      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &timeStart, NULL);
      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &timeEnd, NULL);

      ss.str("");
      ss << ((int)(timeEnd - timeStart) / 1000);

      log->writeLine(CLDebug, ss.str() + " ms");

      clReleaseEvent(events[i]);
    }

    delete[] events;
  }
  else if (waitToFinish) {
    status = clWaitForEvents(bufferCount, events);

    for (unsigned int i = 0; i < bufferCount; i++) {
      clReleaseEvent(events[i]);
    }

    delete[] events;

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while waiting for events: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  hostDataChanged = false;
  hostHasData = false;

  return true;
}

bool CLGlobalBuffer::copyFrom(CLGlobalBuffer* var, bool waitToFinish) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Copying variable: " + var->getName() + ", to variable: " + name);

  if (var == NULL) {
    log->writeLine(CLError, "Cannot copy from NULL variable");
    return false;
  }

  if (memorySize == 0 || link == NULL || clBuffers == NULL) {
    log->writeLine(CLError, "Cannot write to uninitialized OpenCL buffer");
    return false;
  }

  cl_int status;
  cl_event* events = NULL;

  if (CLSystem::getInstance()->getProfiling() || waitToFinish) {
    events = new cl_event[bufferCount];
  }

  for (unsigned int i = 0; i < bufferCount; i++) {
    if (events != NULL) {
      status = clEnqueueCopyBuffer(link->getQueue(i), var->clBuffers[i], clBuffers[i], 0, 0, var->getMemorySize(), 0, NULL, &events[i]);
    }
    else {
      status = clEnqueueCopyBuffer(link->getQueue(i), var->clBuffers[i], clBuffers[i], 0, 0, var->getMemorySize(), 0, NULL, NULL);
    }

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while enqueing copy buffer: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  if (CLSystem::getInstance()->getProfiling()) {
    clWaitForEvents(bufferCount, events);

    std::ostringstream ss;

    for (unsigned int i = 0; i < bufferCount; i++) {
      cl_ulong timeStart, timeEnd;

      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &timeStart, NULL);
      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &timeEnd, NULL);

      ss.str("");
      ss << ((int)(timeEnd - timeStart) / 1000);

      log->writeLine(CLDebug, ss.str() + " ms");

      clReleaseEvent(events[i]);
    }

    delete[] events;
  }
  else if (waitToFinish) {
    status = clWaitForEvents(bufferCount, events);

    for (unsigned int i = 0; i < bufferCount; i++) {
      clReleaseEvent(events[i]);
    }

    delete[] events;

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while waiting for events: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  hostHasData = false;
  hostDataChanged = false;

  return true;
}

bool CLGlobalBuffer::allocateHostData() {
  data = new char[getMemorySize()]();
  hostHasData = false;
  hostDataChanged = false;

  return data != NULL ? true : false;
}

bool CLGlobalBuffer::allocate() {
  Log* log = Log::getInstance();

  bool allocateHost = data != NULL ? false : true;

  release();

  log->writeLine(CLDebug, "Allocating on device buffer: " + name);

  if (memorySize == 0) {
    log->writeLine(CLError, "Cannot allocate OpenCL Global buffer variable without parameters set");
    return false;
  }

  if (link == NULL) {
    log->writeLine(CLError, "Cannot allocate OpenCL Global buffer variable without link");
    return false;
  }

  cl_mem_flags flag = CL_MEM_READ_WRITE;
  bufferCount = link->getDeviceCount();
  //bufferCount = 1; 

  cl_int status;

  if (isSplit()) {
    clBuffers = new cl_mem[bufferCount];
    partElementCount = new size_t[bufferCount];
    offsets = new size_t[bufferCount];
    size_t off = 0;

    for (unsigned int i = 0; i < bufferCount; i++) {
      offsets[i] = off;
      partElementCount[i] = (size_t)ceil((1.0 / bufferCount) * elementCount);

      off += partElementCount[i];
      if (off > elementCount) {
        partElementCount[i] -= off - elementCount;
      }

      clBuffers[i] = clCreateBuffer(link->getContext(), flag, partElementCount[i] * CLSystem::dataTypeSize(varDataType), NULL, &status);

      if (status != CL_SUCCESS) {
        log->writeLine(CLError, "Error while creating Global buffer: " + CLSystem::statusDesc(status));
        return false;
      }
    }
  }
  else {
    clBuffers = new cl_mem;
    clBuffers[0] = clCreateBuffer(link->getContext(), flag, memorySize, NULL, &status);

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while creating Global buffer: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  if (allocateHost) {
    allocateHostData();
  }

  needsUpdate = false;

  return true;
}

void CLGlobalBuffer::release() {
  Log* log = Log::getInstance();

  cl_int status;

  if (clBuffers != NULL) {
    if (isSplit()) {
      for (unsigned int i = 0; i < bufferCount; i++) {
        status = clReleaseMemObject(clBuffers[i]);

        if (status != CL_SUCCESS) {
          log->writeLine(CLError, "Releasing global buffer " + name + " failed: " + CLSystem::statusDesc(status));
        }
      }
    }
    else {
      status = clReleaseMemObject(clBuffers[0]);

      if (status != CL_SUCCESS) {
        log->writeLine(CLError, "Releasing global buffer " + name + " failed: " + CLSystem::statusDesc(status));
      }
    }

    delete[] clBuffers;
  }

  if (partElementCount > 0) {
    delete[] partElementCount;
  }

  if (offsets > 0) {
    delete[] offsets;
  }

  bufferCount = 0;
  clBuffers = NULL;
  partElementCount = NULL;
  offsets = NULL;

  if (data != NULL) {
    delete[] data;
    data = NULL;
  }

  hostDataChanged = false;
  hostHasData = false;
}

bool CLGlobalBuffer::setAsArgument(CLProgram* program, unsigned int argID, unsigned int deviceID, size_t kernelLocalSize) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Setting global buffer kernel argument " + name);

  if (program == NULL) {
    log->writeLine(CLError, "Cannot set global buffer kernel argument to NULL program");
    return false;
  }

  if (program->getKernel() == NULL) {
    log->writeLine(CLError, "Cannot set global buffer kernel argument to NULL kernel");
    return false;
  }

  cl_int status;

  if (isSplit()) {
    status = clSetKernelArg(program->getKernel(), (cl_uint)argID, sizeof(cl_mem), &clBuffers[deviceID]);
  }
  else {
    status = clSetKernelArg(program->getKernel(), (cl_uint)argID, sizeof(cl_mem), clBuffers);
  }

  if (status != CL_SUCCESS) {
    log->writeLine(CLError, "Error while setting Global buffer kernel argument " + name + ": " + CLSystem::statusDesc(status));
    return false;
  }

  return true;
}

void CLGlobalBuffer::reallocate(unsigned int elements) {
  release();
  this->elementCount = elements;
  this->memorySize = elements * this->varDataTypeSize;
  allocate();
}

double CLGlobalBuffer::getScalar(unsigned int id) {
  if (!hostHasData) {
    if (!download()) {
      return DBL_MAX;
    }
  }

  return CLVariable::getScalar(id);
}

Vector<3, double> CLGlobalBuffer::getVector(unsigned int id) {
  if (!hostHasData) {
    if (!download()) {
      return DBL_MAX;
    }
  }

  return CLVariable::getVector(id);
}

bool CLGlobalBuffer::setScalar(unsigned int id, double var) {
  if (data == NULL) {
    if (!allocateHostData()) {
      return false;
    }
  }

  if (!hostHasData) {
    if (!download()) {
      return false;
    }
  }

  if (!CLVariable::setScalar(id, var)) {
    return false;
  }

  hostDataChanged = true;

  return true;
}

bool CLGlobalBuffer::setVector(unsigned int id, Vector<3, double> var) {
  if (data == NULL) {
    if (!allocateHostData()) {
      return false;
    }
  }

  if (!hostHasData) {
    if (!download()) {
      return false;
    }
  }

  if (!CLVariable::setVector(id, var)) {
    return false;
  }

  hostDataChanged = true;

  return true;
}