#include "clframework.h"

using namespace clframework;

CLLink::CLLink(std::vector<CLDevice*>& devices):
  context(NULL) {

  initialize(devices);
}

CLLink::CLLink(CLDevice* device):
  context(NULL) {

  if (device != NULL) {
    std::vector<CLDevice*> devices;
    devices.push_back(device);

    initialize(devices);
  }
  else {
    Log::getInstance()->writeLine(CLError, "Cannot connect to NULL device!");
  }
}

CLLink::CLLink(CLPlatform* platform):
  context(NULL) {

  if (platform == NULL) {
    platform = CLSystem::getInstance()->getFirstPlatform();
  }

  if (platform == NULL) {
    Log::getInstance()->writeLine(CLError, "There are no platform to connect to");
    return;
  }

  std::vector<CLDevice*> devices = platform->getDevices();
  initialize(devices);
}

CLLink::~CLLink() {
  Log::getInstance()->writeLine(CLDebug, "Releasing link to devices");

  cl_int status;

  for (cl_uint i = 0; i < queues.size(); i++) {
    status = clReleaseCommandQueue(queues[i]);
  }

  devices.clear();
  queues.clear();

  if (context != NULL) {
    status = clReleaseContext(context);
  }
}

bool CLLink::initialize(std::vector<CLDevice*>& devices) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Connecting to OpenCL enabled device(s)");

  if (devices.empty()) {
    log->writeLine(CLError, "No devices to connect to");
    return false;
  }

  cl_device_id *devicesIDs = new cl_device_id[devices.size()];

  for (cl_uint i = 0; i < devices.size(); i++) {
    this->devices.push_back(devices[i]);
    devicesIDs[i] = devices[i]->getId();
  }

  cl_int status;
  context = clCreateContext(NULL, (cl_uint)devices.size(), devicesIDs, NULL, NULL, &status);

  if (status != CL_SUCCESS) {
    log->writeLine(CLError, "Error while creating context: " + CLSystem::statusDesc(status));
    return false;
  }

  for (cl_uint i = 0; i < devices.size(); i++) {
    log->writeLine(CLDebug, "Linking to device: " + devices[i]->getName());

    if (CLSystem::getInstance()->getProfiling()) {
      queues.push_back(clCreateCommandQueue(context, devices[i]->getId(), CL_QUEUE_PROFILING_ENABLE, &status));
    }
    else {
      queues.push_back(clCreateCommandQueue(context, devices[i]->getId(), 0, &status));
    }

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while creating command queue: " + CLSystem::statusDesc(status));
      delete[] devicesIDs;
      return false;
    }
  }

  delete[] devicesIDs;

  return true;
}

bool CLLink::buildProgram(cl_program program, const std::string& buildOptions) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Building program with associated link");

  for (unsigned int i = 0; i < devices.size(); i++) {
    std::string finalBuildOptions = buildOptions;

    finalBuildOptions.append(" -cl-mad-enable");

    if (devices[i]->isGPU()) {
      finalBuildOptions.append(" -D GPU");

      if (getPlatform()->getName().find("ATI") != std::string::npos) {
        finalBuildOptions.append(" -D ATI");
      }
      else {
        finalBuildOptions.append(" -D NVIDIA");
      }
    }
    else if (devices[i]->isCPU()) {
      finalBuildOptions.append(" -D CPU");
    }

    cl_int status = clBuildProgram(program, 1, &devices[i]->getId(), finalBuildOptions.c_str(), NULL, NULL);

    const size_t size = 4096 * 8;
    char *buf = new char[size];

    cl_int status2 = clGetProgramBuildInfo(program, devices[i]->getId(), CL_PROGRAM_BUILD_LOG, size, buf, NULL);

    if (status2 != CL_SUCCESS) {
      log->writeLine(CLError, "Error while fetching OpenCL compile log");
    }
    else {
      std::string bufStr(buf);
      log->writeLine(CLDebug, "Building program info: " + ((bufStr == "") ? "OK" : bufStr));
    }

    delete[] buf;

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while building program: " + CLSystem::statusDesc(status));
      return false;
    }
  }

  return true;
}

bool CLLink::finish() {
  cl_int status;
  bool success = true;

  for (unsigned int i = 0; i < queues.size(); i++) {
    status = clFinish(queues[i]);

    if (status != CL_SUCCESS) {
      success = false;
      Log::getInstance()->writeLine(CLError, "Error while finishing program: " + CLSystem::statusDesc(status));
    }
  }

  return success;
}

unsigned int CLLink::getMinMaxWorgroupSize() {
  unsigned int minValue = 100000;

  for (unsigned int i = 0; i < devices.size(); i++) {
    if (devices[i]->getMaxWorkGroupSize() < minValue) {
      minValue = devices[i]->getMaxWorkGroupSize();
    }
  }

  return minValue;
}

CLPlatform* CLLink::getPlatform() {
  return devices.empty() ? NULL : devices[0]->getPlatform();
}