#include "clframework.h"

using namespace clframework;

CLProgram::CLProgram(std::string sourcePath, std::string name):
  link(NULL),
  program(NULL),
  kernel(NULL),
  workGroupSizes(NULL),
  hasDynamicGlobalBuffer(false),
  isBuilt(false),
  unsafeMath(false),
  name(name) {

  kernelSource = new CLFunction(sourcePath, "kernel");
}

CLProgram::CLProgram(std::string sourcePath, std::string name, CLLink* link) :
  link(link),
  program(NULL),
  kernel(NULL),
  workGroupSizes(NULL),
  hasDynamicGlobalBuffer(false),
  isBuilt(false),
  unsafeMath(false),
  name(name) {

  kernelSource = new CLFunction(sourcePath, "kernel");
}

CLProgram::~CLProgram() {
  releaseProgram();
}

bool CLProgram::createKernel() {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Creating kernel for program: " + name);

  if (program == NULL) {
    log->writeLine(CLError, "Program wasn't build yet");
    return false;
  }

  if (link == NULL) {
    log->writeLine(CLError, "No link to devices set to create kernel for");
    return false;
  }

  cl_int status;

  releaseKernel();
  kernel = clCreateKernel(program, name.c_str(), &status);

  if (status != CL_SUCCESS) {
    log->writeLine(CLError, "Error while creating kernel: " + CLSystem::statusDesc(status));
    return false;
  }

  if (!setPersistentArguments()) {
    return false;
  }

  workGroupSizes = new size_t[link->getDeviceCount()];

  for (unsigned int i = 0; i < link->getDeviceCount(); i++) {
    status = clGetKernelWorkGroupInfo(kernel, link->getDevice(i)->getId(), CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &workGroupSizes[i], NULL);

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while getting kernel work group info: " + CLSystem::statusDesc(status));
      return false;
    }

    workGroupSizes[i] = (std::min)(workGroupSizes[i], (size_t)1024);
  }

  return true;
}

void CLProgram::releaseKernel() {
  Log::getInstance()->writeLine(CLDebug, "Releasing kernel " + name);

  if (kernel != NULL) {
    clReleaseKernel(kernel);
    kernel = NULL;
  }

  if (workGroupSizes > 0) {
    delete[] workGroupSizes;
    workGroupSizes = NULL;
  }
}

void CLProgram::releaseProgram() {
  Log::getInstance()->writeLine(CLDebug, "Releasing program " + name);

  releaseKernel();

  /*for (std::map<std::string, CLVariable*>::iterator it = variables.begin(); it != variables.end(); it++) {
    CLVariable* var = it->second;

    if (var != NULL) {
      delete var;
    }
  }*/

  functionsSources.clear();
  variablesArgIDs.clear();
  variables.clear();
  globalBuffers.clear();
  localBuffers.clear();
  kernelArguments.clear();
  constants.clear();

  if (program != NULL) {
    cl_int status = clReleaseProgram(program);
  }
}

bool CLProgram::setPersistentArguments() {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Setting persistent arguments to kernel");

  if (kernel == NULL) {
    log->writeLine(CLError, "Cannot set persistent arguments to NULL kernel");
    return false;
  }

  for (std::map<std::string, CLVariable*>::iterator it = variables.begin(); it != variables.end(); it++) {
    CLVariable* var = it->second;

    if (var != NULL) {
      if (var->getVarType() == GlobalBuffer) {
        if (!var->setAsArgument(this, variablesArgIDs.find(var)->second)) {
            return false;
        }
      }
    }
  }
  
  return true;
}

void CLProgram::setUnsafeMath(bool value) {
  unsafeMath = value;
  isBuilt = false;
}

void CLProgram::addBuildOption(const std::string& option) {
  buildOptions.push_back(option);
  isBuilt = false;
}

void CLProgram::clearBuildOptions() {
  buildOptions.clear();
  isBuilt = false;
}

void CLProgram::addFunction(CLFunction* function) {
  if (function == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add NULL function " + name);
    return;
  }

  functionsSources.insert(std::pair<std::string, CLFunction*>(function->getName(), function));
}

void CLProgram::addKernelArgument(CLKernelArgument* kernelArgument, unsigned int argID) {
  if (kernelArgument == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add NULL kernel argument to program " + name);
    return;
  }

  variables.insert(std::pair<std::string, CLVariable*>(kernelArgument->getName(), kernelArgument));
  kernelArguments.insert(std::pair<std::string, CLKernelArgument*>(kernelArgument->getName(), kernelArgument));

  variablesArgIDs.insert(std::pair<CLVariable*, unsigned int>(kernelArgument, argID));
}

void CLProgram::addGlobalBuffer(CLGlobalBuffer* globalBuffer, unsigned int argID) {
  if (globalBuffer == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add NULL global buffer to program " + name);
    return;
  }
  
  if (globalBuffer->getLink() == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add global buffer with NULL link set to program");
    return;
  }

  if (globalBuffer->getLink() != link) {
    Log::getInstance()->writeLine(CLError, "Cannot add global buffer to program with different link set");
    return;
  }

  if (globalBuffer->getIsDynamic()) {
    hasDynamicGlobalBuffer = true;
  }

  variables.insert(std::pair<std::string, CLVariable*>(globalBuffer->getName(), globalBuffer));
  globalBuffers.insert(std::pair<std::string, CLGlobalBuffer*>(globalBuffer->getName(), globalBuffer));

  variablesArgIDs.insert(std::pair<CLVariable*, unsigned int>(globalBuffer, argID));
}

void CLProgram::addLocalBuffer(CLLocalBuffer* localBuffer, unsigned int argID) {
  if (localBuffer == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add NULL local buffer to program " + name);
    return;
  }

  variables.insert(std::pair<std::string, CLVariable*>(localBuffer->getName(), localBuffer));
  localBuffers.insert(std::pair<std::string, CLLocalBuffer*>(localBuffer->getName(), localBuffer));

  variablesArgIDs.insert(std::pair<CLVariable*, unsigned int>(localBuffer, argID));
}

void CLProgram::addConstant(CLConstant* constant) {
  if (constant == NULL) {
    Log::getInstance()->writeLine(CLError, "Cannot add NULL constant to program " + name);
    return;
  }

  constants.insert(std::pair<std::string, CLConstant*>(constant->getName(), constant));
}

CLKernelArgument* CLProgram::getKernelArgument(std::string name) {
  std::map<std::string, CLKernelArgument*>::iterator found = kernelArguments.find(name);

  if (found != kernelArguments.end()) {
    return found->second;
  }

  return NULL;
}

CLGlobalBuffer* CLProgram::getGlobalBuffer(std::string name) {
  std::map<std::string, CLGlobalBuffer*>::iterator found = globalBuffers.find(name);

  if (found != globalBuffers.end()) {
    return found->second;
  }

  return NULL;
}

CLLocalBuffer* CLProgram::getLocalBuffer(std::string name) {
  std::map<std::string, CLLocalBuffer*>::iterator found = localBuffers.find(name);

  if (found != localBuffers.end()) {
    return found->second;
  }

  return NULL;
}

bool CLProgram::setLink(CLLink* linkToDevices) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Setting new link to devices for program " + name);

  if (linkToDevices == NULL) {
    log->writeLine(CLError, "Cannot set NULL link to devices");
    return false;
  }

  if (linkToDevices->getDeviceCount() == 0) {
    log->writeLine(CLError, "Link isn't connected to any device");
    return false;
  }

  link = linkToDevices;
  isBuilt = false;

  return true;
}

bool CLProgram::build() {
  Log* log = Log::getInstance();

  if (isBuilt) {
    releaseProgram();
  }

  log->writeLine(CLDebug, "Building program " + name);
  isBuilt = false;

  if (link == NULL) {
    log->writeLine(CLError, "No OpenCL devices set to build program on");
    return false;
  }

  std::string sourceFinal = "";

  for (std::map<std::string, CLConstant*>::iterator it = constants.begin(); it != constants.end(); it++) {
    sourceFinal.append("#define " + it->first + " (" + it->second->getSource() + ")\n");
  }

  for (std::map<std::string, CLFunction*>::iterator it = functionsSources.begin(); it != functionsSources.end(); it++) {
    sourceFinal.append(it->second->getSource() + "\n");
  }

  sourceFinal.append(kernelSource->getSource());

  cl_int status;

  const char* sourceCstr = sourceFinal.c_str();
  size_t sourceSize = sourceFinal.size();
  program = clCreateProgramWithSource(link->getContext(), 1, &sourceCstr, &sourceSize, &status);

  if (status != CL_SUCCESS) {
    log->writeLine(CLError, "Error while creating program with source : " + CLSystem::statusDesc(status));
    return false;
  }

  std::string buildOptionsStr;
  if (unsafeMath) {
    buildOptionsStr = "-cl-fast-relaxed-math";
  }
  else {
    buildOptionsStr = "-cl-finite-math-only";
  }

  for (unsigned int i = 0; i < buildOptions.size(); i++) {
    buildOptionsStr.append(' ' + buildOptions[i]);
  }

  if (!link->buildProgram(program, buildOptionsStr)) {
    return false;
  }

  for (std::map<std::string, CLVariable*>::iterator it = variables.begin(); it != variables.end(); it++) {
    if (it->second->getNeedsUpdate()) {
      if (!it->second->allocate()) {
        return false;
      }
    }
  }

  if (!createKernel()) {
    return false;
  }

  isBuilt = true;

  return isBuilt;
}

bool CLProgram::enqueue(size_t globalSize, size_t localSize) {
  Log* log = Log::getInstance();

  log->writeLine(CLDebug, "Enqueuing kernel for program " + name);

  if (!isBuilt) {
    log->writeLine(CLError, "Cannot run kernel of not built program");
    return false;
  }

  if (link == NULL) {
    log->writeLine(CLError, "No link to devices set to run kernel on");
    return false;
  }

  if (localSize > globalSize) {
    log->writeLine(CLError, "Local size of program can't be greater than the global size");
    return false;
  }

  cl_int status;
  cl_event* events = NULL;

  if (CLSystem::getInstance()->getProfiling()) {
    events = new cl_event[link->getDeviceCount()];
  }

  for (unsigned int i = 0; i < link->getDeviceCount(); i++) {
    CLVariable *var = NULL;

    if (globalSize == 0) {
      for (std::map<std::string, CLVariable*>::iterator it = variables.begin(); it != variables.end(); it++) {
        if (it->second->getParalelize()) {
          var = it->second;
          break;
        }
      }

      if (var) {
        globalSize = var->getElementCount(i);
      }
      else {
        log->writeLine(CLError, "Error choosing global size for kernel " + name);
        return false;
      }
    }

    if (localSize == 0) {
      localSize = workGroupSizes[i];
    }

    while (globalSize % localSize != 0) localSize--;

    for (std::map<std::string, CLVariable*>::iterator it = variables.begin(); it != variables.end(); it++) {
      var = it->second;

      if (var != NULL) {
        if (var->getVarType() != GlobalBuffer || ((CLGlobalBuffer*)var)->getIsDynamic()) {
          if (!var->setAsArgument(this, variablesArgIDs.find(var)->second, i, localSize)) {
            return false;
          }
        }
      }
      else {
        log->writeLine(CLError, "Program doesn't contain variable needed by kernel " + name);
        return false;
      }
    }

    if (events) {
      status = clEnqueueNDRangeKernel(link->getQueue(i), kernel, 1, NULL, &globalSize, &localSize, 0, NULL, &events[i]);
    }
    else {
      status = clEnqueueNDRangeKernel(link->getQueue(i), kernel, 1, NULL, &globalSize, &localSize, 0, NULL, NULL);
    }

    if (status != CL_SUCCESS) {
      log->writeLine(CLError, "Error while enqueing kernel " + name + ": " + CLSystem::statusDesc(status));
      return false;
    }
  }

  if (CLSystem::getInstance()->getProfiling()) {
    clWaitForEvents(link->getDeviceCount(), events);

    std::ostringstream ss;

    for (unsigned int i = 0; i < link->getDeviceCount(); i++) {
      cl_ulong timeStart, timeEnd;

      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &timeStart, NULL);
      clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &timeEnd, NULL);

      ss.str("");
      ss << ((int)(timeEnd - timeStart) / 1000);

      log->writeLine(CLDebug, ss.str() + " ms");

      clReleaseEvent(events[i]);
    }

    delete[] events;
  }

  return true;
}

bool CLProgram::finish() {
  if (!isBuilt) {
    return false;
  }

  if (!link) {
    return false;
  }

  return link->finish();
}

std::string CLProgram::getCompiledBinary() {
  if (!isBuilt) {
    return "";
  }

  size_t binarySize;
  cl_int status = clGetProgramInfo(program, CL_PROGRAM_BINARY_SIZES, sizeof(size_t), &binarySize, NULL);
  char *binary = new char[binarySize + 1];
  status = clGetProgramInfo(program, CL_PROGRAM_BINARIES, sizeof(char*), &binary, NULL);

  return binary;
}