#include "clframework.h"

using namespace clframework;

CLSystem* CLSystem::clsystemInstance = NULL;

CLSystem::CLSystem() {
  initialize();
}

CLSystem::~CLSystem() {
  Log::getInstance()->writeLine(CLDebug, "Unloading OpenCL system");

  for (unsigned int i = 0; i < platforms.size(); i++) {
    delete platforms[i];
  }

  platforms.clear();

  clsystemInstance = NULL;
}

void CLSystem::initialize() {
  Log* log = Log::getInstance(); 
  log->writeLine(CLInfo, "Initializing OpenCL system");
  
  profiling = false;

  cl_platform_id platformIDs[10];
  cl_uint platformCount;
  cl_int status = clGetPlatformIDs(10, platformIDs, &platformCount);

  for (unsigned int i = 0; i < platformCount; i++) {
    CLPlatform* platform = new CLPlatform(platformIDs[i]);
    platform->initialize();

    platforms.push_back(platform);
  }
}

std::string CLSystem::statusDesc(cl_int status) {
  switch (status) {
    case CL_SUCCESS:						                return "OpenCL - Success!";
    case CL_DEVICE_NOT_FOUND:				            return "OpenCL - Device not found.";
    case CL_DEVICE_NOT_AVAILABLE:			          return "OpenCL - Device not available";
    case CL_COMPILER_NOT_AVAILABLE:			        return "OpenCL - Compiler not available";
    case CL_MEM_OBJECT_ALLOCATION_FAILURE:	    return "OpenCL - Memory object allocation failure";
    case CL_OUT_OF_RESOURCES:				            return "OpenCL - Out of resources";
    case CL_OUT_OF_HOST_MEMORY:				          return "OpenCL - Out of host memory";
    case CL_PROFILING_INFO_NOT_AVAILABLE:	      return "OpenCL - Profiling information not available";
    case CL_MEM_COPY_OVERLAP:                   return "OpenCL - Memory copy overlap";
    case CL_IMAGE_FORMAT_MISMATCH:              return "OpenCL - Image format mismatch";
    case CL_IMAGE_FORMAT_NOT_SUPPORTED:         return "OpenCL - Image format not supported";
    case CL_BUILD_PROGRAM_FAILURE:			        return "OpenCL - Program build failure";
    case CL_MAP_FAILURE:					              return "OpenCL - Map failure";
    case CL_INVALID_VALUE:					            return "OpenCL - Invalid value";
    case CL_INVALID_DEVICE_TYPE:			          return "OpenCL - Invalid device type";
    case CL_INVALID_PLATFORM:				            return "OpenCL - Invalid platform";
    case CL_INVALID_DEVICE:					            return "OpenCL - Invalid device";
    case CL_INVALID_CONTEXT:				            return "OpenCL - Invalid context";
    case CL_INVALID_QUEUE_PROPERTIES:		        return "OpenCL - Invalid queue properties";
    case CL_INVALID_COMMAND_QUEUE:			        return "OpenCL - Invalid command queue";
    case CL_INVALID_HOST_PTR:				            return "OpenCL - Invalid host pointer";
    case CL_INVALID_MEM_OBJECT:				          return "OpenCL - Invalid memory object";
    case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:    return "OpenCL - Invalid image format descriptor";
    case CL_INVALID_IMAGE_SIZE:				          return "OpenCL - Invalid image size";
    case CL_INVALID_SAMPLER:				            return "OpenCL - Invalid sampler";
    case CL_INVALID_BINARY:					            return "OpenCL - Invalid binary";
    case CL_INVALID_BUILD_OPTIONS:			        return "OpenCL - Invalid build options";
    case CL_INVALID_PROGRAM:				            return "OpenCL - Invalid program";
    case CL_INVALID_PROGRAM_EXECUTABLE:		      return "OpenCL - Invalid program executable";
    case CL_INVALID_KERNEL_NAME:			          return "OpenCL - Invalid kernel name";
    case CL_INVALID_KERNEL_DEFINITION:		      return "OpenCL - Invalid kernel definition";
    case CL_INVALID_KERNEL:					            return "OpenCL - Invalid kernel";
    case CL_INVALID_ARG_INDEX:				          return "OpenCL - Invalid argument index";
    case CL_INVALID_ARG_VALUE:				          return "OpenCL - Invalid argument value";
    case CL_INVALID_ARG_SIZE:				            return "OpenCL - Invalid argument size";
    case CL_INVALID_KERNEL_ARGS:			          return "OpenCL - Invalid kernel arguments";
    case CL_INVALID_WORK_DIMENSION:			        return "OpenCL - Invalid work dimension";
    case CL_INVALID_WORK_GROUP_SIZE:		        return "OpenCL - Invalid work group size";
    case CL_INVALID_WORK_ITEM_SIZE:			        return "OpenCL - Invalid work item size";
    case CL_INVALID_GLOBAL_OFFSET:			        return "OpenCL - Invalid global offset";
    case CL_INVALID_EVENT_WAIT_LIST:		        return "OpenCL - Invalid event wait list";
    case CL_INVALID_EVENT:					            return "OpenCL - Invalid event";
    case CL_INVALID_OPERATION:				          return "OpenCL - Invalid operation";
    case CL_INVALID_GL_OBJECT:				          return "OpenCL - Invalid OpenGL object";
    case CL_INVALID_BUFFER_SIZE:			          return "OpenCL - Invalid buffer size";
    case CL_INVALID_MIP_LEVEL:				          return "OpenCL - Invalid mip-map level";
    default:								                    return "OpenCL - Unknown error";
  }
}

std::string CLSystem::deviceTypeString(cl_device_type type) {
  switch (type) {
    case CL_DEVICE_TYPE_ACCELERATOR:          return "CL_DEVICE_TYPE_ACCELERATOR";
    case CL_DEVICE_TYPE_ALL:                  return "CL_DEVICE_TYPE_ALL";
    case CL_DEVICE_TYPE_CPU:                  return "CL_DEVICE_TYPE_CPU";
    case CL_DEVICE_TYPE_CUSTOM:               return "CL_DEVICE_TYPE_CUSTOM";
    case CL_DEVICE_TYPE_DEFAULT:              return "CL_DEVICE_TYPE_DEFAULT";
    case CL_DEVICE_TYPE_GPU:                  return "CL_DEVICE_TYPE_GPU";
    default:                                  return "CL_INVALID_DEVICE_TYPE";
  }
}

std::string CLSystem::deviceTypeStringShort(cl_device_type type) {
  switch (type) {
    case CL_DEVICE_TYPE_ACCELERATOR:          return "ACCELERATOR";
    case CL_DEVICE_TYPE_ALL:                  return "ALL";
    case CL_DEVICE_TYPE_CPU:                  return "CPU";
    case CL_DEVICE_TYPE_CUSTOM:               return "CUSTOM";
    case CL_DEVICE_TYPE_DEFAULT:              return "DEFAULT";
    case CL_DEVICE_TYPE_GPU:                  return "GPU";
    default:                                  return "CL_INVALID_DEVICE_TYPE";
  }
}

size_t CLSystem::dataTypeSize(CLVariableDataType type) {
  switch (type) {
    case FloatType:                             return sizeof(cl_float);
    case Float2Type:                            return sizeof(cl_float2);
    case Float3Type:                            return sizeof(cl_float3);
    case Float4Type:                            return sizeof(cl_float4);
    case DoubleType:                            return sizeof(cl_double);
    case Double2Type:                           return sizeof(cl_double2);
    case Double3Type:                           return sizeof(cl_double3);
    case Double4Type:                           return sizeof(cl_double4);
    case UIntType:                              return sizeof(cl_uint);
    case UInt2Type:                             return sizeof(cl_uint2);
    case UInt3Type:                             return sizeof(cl_uint3);
    case UInt4Type:                             return sizeof(cl_uint4);
    case LongType:                              return sizeof(cl_long);
    case ULongType:                             return sizeof(cl_ulong);
    case Long2Type:                             return sizeof(cl_long2);
    case ULong2Type:                            return sizeof(cl_ulong2);
    case IntType:                               return sizeof(cl_int);
    case Int2Type:                              return sizeof(cl_int2);
    case Int3Type:                              return sizeof(cl_int3);
    case Int4Type:                              return sizeof(cl_int4);
    case CharType:                              return sizeof(cl_char);
    case UCharType:                             return sizeof(cl_uchar);
    default:                                    return 0;
  }
}

CLSystem* CLSystem::getInstance() {
  if (clsystemInstance == NULL) {
    clsystemInstance = new CLSystem();
  }

  return clsystemInstance;
}