#include "clframework.h"

using namespace clframework;

CLVariable::CLVariable(std::string name, CLVariableType varType, CLVariableDataType varDataType, unsigned int elements, bool paralelize):
  name(name),
  varType(varType),
  varDataType(varDataType),
  data(NULL),
  elementCount(0),
  memorySize(0),
  paralelize(paralelize) {

  this->elementCount = elements;
  this->varDataType = varDataType;
  this->varDataTypeSize = CLSystem::dataTypeSize(varDataType);
  this->memorySize = elements * this->varDataTypeSize;
}

CLVariable::~CLVariable() {
  Log::getInstance()->writeLine(CLDebug, "Destroying variable " + name);
}

void* CLVariable::get(unsigned int id) {
  return (void*)(data + varDataTypeSize * id);
}

double CLVariable::getScalar(unsigned int id) {
  if (data == NULL) {
    return DBL_MAX;
  }

  switch (varDataType) {
    case DoubleType:	      return *(cl_double*)get(id);
    case FloatType:		      return *(cl_float*)get(id);
    case LongType:		      return *(cl_long*)get(id);
    case ULongType:		      return *(cl_ulong*)get(id);
    case IntType:		        return *(cl_int*)get(id);
    case UIntType:		      return *(cl_uint*)get(id);
    case CharType:		      return *(cl_char*)get(id);
    case UCharType:		      return *(cl_uchar*)get(id);
    default:                Log::getInstance()->writeLine(CLError, "Trying to read scalar value from some different data type");
                            return DBL_MAX;
  }
}

Vector<3, double> CLVariable::getVector(unsigned int id) {
  if (data == NULL) {
    return Vector<3, double>(DBL_MAX);
  }

  switch (varDataType) {
    case Float2Type:	      return *(Vector<2, float>*)get(id);
    case Float3Type:	      return *(Vector<3, float>*)get(id);
    case Float4Type:	      return *(Vector<3, float>*)get(id);
    case Double2Type:	      return *(Vector<2, double>*)get(id);
    case Double3Type:	      return *(Vector<3, double>*)get(id);
    case Double4Type:	      return *(Vector<3, double>*)get(id);
    case Int2Type:		      return *(Vector<2, int>*)get(id);
    case Int3Type:		      return *(Vector<3, int>*)get(id);
    case Int4Type:		      return *(Vector<3, int>*)get(id);
    case UInt2Type:		      return *(Vector<2, unsigned int>*)get(id);
    case UInt3Type:		      return *(Vector<3, unsigned int>*)get(id);
    case UInt4Type:		      return *(Vector<3, unsigned int>*)get(id);
    case Long2Type:		      return *(Vector<2, long long>*)get(id);
    case ULong2Type:		    return *(Vector<2, unsigned long long>*)get(id);
    default:                Log::getInstance()->writeLine(CLError, "Trying to read Vector value from some different data type");
                            return Vector<3, double>(DBL_MAX);
  }
}

bool CLVariable::setScalar(unsigned int id, double var) {
  if (data == NULL) {
    return false;
  }

  void* i = get(id);

  switch (varDataType) {
    case DoubleType:	      *(cl_double*)i = var; break;
    case FloatType:		      *(cl_float*)i = (cl_float)var; break;
    case LongType:		      *(cl_long*)i = (cl_long)var; break;
    case ULongType:		      *(cl_ulong*)i = (cl_ulong)var; break;
    case IntType:		        *(cl_int*)i = (cl_int)var; break;
    case UIntType:		      *(cl_uint*)i = (cl_uint)var; break;
    case CharType:		      *(cl_char*)i = (cl_char)var; break;
    case UCharType:		      *(cl_uchar*)i = (cl_uchar)var; break;
    default:                Log::getInstance()->writeLine(CLError, "Trying to write scalar value to some different data type");
                            return false;
  }

  return true;
}

bool CLVariable::setVector(unsigned int id, Vector<3, double> var) {
  if (data == NULL) {
    return false;
  }

  void* i = get(id);

  switch (varDataType) {
    case Float2Type:	        *(Vector<2, float>*)i = var; break;
    case Float3Type:	        *(Vector<3, float>*)i = var; break;
    case Float4Type:	        *(Vector<3, float>*)i = var; *((float*)i + 3) = 0.0f; break;
    case Double2Type:	        *(Vector<2, double>*)i = var; break;
    case Double3Type:	        *(Vector<3, double>*)i = var; break;
    case Double4Type:	        *(Vector<3, double>*)i = var; *((double*)i + 3) = 0.0; break;
    case Int2Type:		        *(Vector<2, int>*)i = var; break;
    case Int3Type:		        *(Vector<3, int>*)i = var; break;
    case Int4Type:		        *(Vector<3, int>*)i = var; *((int*)i + 3) = 0; break;
    case UInt2Type:		        *(Vector<2, unsigned int>*)i = var; break;
    case UInt3Type:		        *(Vector<3, unsigned int>*)i = var; break;
    case UInt4Type:		        *(Vector<3, unsigned int>*)i = var; *((unsigned int*)i + 3) = 0; break;
    case Long2Type:		        *(Vector<2, long long>*)i = var; break;
    case ULong2Type:		      *(Vector<2, unsigned long long>*)i = var; break;
    default:                  Log::getInstance()->writeLine(CLError, "Trying to write vector value to some different data type");
                              return false;
  }

  return true;
}

bool CLVariable::isScalar() {
  switch (varDataType) {
    case FloatType:
    case DoubleType:
    case LongType:
    case ULongType:
    case IntType:
    case UIntType:
    case CharType:
    case UCharType:
      return true;
    default:
      return false;
  }
}