#include "../main.h"

Camera::Camera(OpenGL* ogl): 
  opengl(ogl) {

  position[0] = 0.0; position[1] = 0.0; position[2] = 0.0; 
  lookAt[0] = 0.0; lookAt[1] = 0.0; lookAt[2] = 0.0; 
  up[0] = 0.0; up[1] = 0.0; up[2] = 1.0; 
}

Camera::Camera(OpenGL* ogl, glm::vec3& position, glm::vec3& lookAt, glm::vec3& up):
  opengl(ogl),
  position(position), 
  lookAt(lookAt), 
  up(up) {

}

GLvoid Camera::useCamera() {
  gluLookAt(position.x, position.y, position.z, lookAt.x, lookAt.y, lookAt.z, up.x, up.y, up.z);
}

glm::mat4 Camera::getViewMatrix() {
  return glm::lookAt(position, lookAt, up);
}


OrbitCamera::OrbitCamera(OpenGL* ogl, GLfloat azimuth, GLfloat altitude, GLfloat distance) :
  Camera(ogl),
  azimuth(azimuth), 
  altitude(altitude),
  distance(distance),
  wasMouseMove(false),
  freeCamera(false) {

  computePosition();
}

OrbitCamera::OrbitCamera(OpenGL* ogl, GLfloat azimuth, GLfloat altitude, glm::vec3 position, GLfloat moveSpeed) :
  Camera(ogl),
  azimuth(azimuth),
  altitude(altitude),
  moveSpeed(moveSpeed),
  distance(0.01f),
  wasMouseMove(false),
  freeCamera(true) {

  lookAt = position;

  computePosition();
}

GLvoid OrbitCamera::computePosition() {
  position.x = lookAt.x + (distance * (GLfloat)cos(azimuth) * (GLfloat)cos(altitude));
	position.y = lookAt.y + (distance * (GLfloat)sin(azimuth) * (GLfloat)cos(altitude));
	position.z = lookAt.z + (distance * (GLfloat)sin(altitude));
}

GLvoid OrbitCamera::moveCamera() {
  Window* window = opengl->getWindow();

  glm::vec3 moveTo(0.0, 0.0, 0.0);
  glm::vec3 moveSide(0.0, 0.0, 0.0);
  glm::vec3 moveHeight(0.0, 0.0, 0.0);
  
  int i = 0, j = 0, k = 0;

  if (window->getKeyAsync(VK_UP) && !window->getKeyAsync(VK_DOWN)) i = 1;
  else if (window->getKeyAsync(VK_DOWN) && !window->getKeyAsync(VK_UP)) i = -1;
  
  if (window->getKeyAsync(VK_LEFT) && !window->getKeyAsync(VK_RIGHT)) j = 1;
  else if (window->getKeyAsync(VK_RIGHT) && !window->getKeyAsync(VK_LEFT)) j = -1;

  if (window->getKeyAsync(VK_SHIFT) && !window->getKeyAsync(VK_CONTROL)) k = 1;
  else if (window->getKeyAsync(VK_CONTROL) && !window->getKeyAsync(VK_SHIFT)) k = -1;

  if (i != 0) {
    moveTo = glm::vec3(i * moveSpeed) * glm::normalize(lookAt - position);
  }
  if (j != 0) {
    glm::vec3 forward = glm::normalize(lookAt - position);
    moveSide = glm::vec3(j * moveSpeed) * glm::normalize(glm::cross(up, forward));
  }
  if (k != 0) {
    glm::vec3 forward = glm::normalize(lookAt - position);
    glm::vec3 side = glm::normalize(glm::cross(forward, up));
    moveHeight = glm::vec3(k * moveSpeed) * glm::normalize(glm::cross(side, forward));
  }

  lookAt += moveTo + moveSide + moveHeight;
  position += moveTo + moveSide + moveHeight;
}

GLvoid OrbitCamera::useCamera() {
  if (freeCamera) {
    moveCamera();
  }
  Camera::useCamera();
}

GLvoid OrbitCamera::keyDown(GLint key) {
  
}

GLvoid OrbitCamera::keyUp(GLint key) {
  
}

GLvoid OrbitCamera::mouseDown(GLint mouseButton, GLint x, GLint y) {
  if (mouseButton & MK_RBUTTON) {
    azimuthOld = azimuth; 
    altitudeOld = altitude; 
  }

  wasMouseMove = false;
}

GLvoid OrbitCamera::mouseUp(GLint mouseButton, GLint x, GLint y) {
  /*if (!freeCamera) {
    if ((mouseButton & MK_LBUTTON) && !wasMouseMove) {
      lookAt = OpenGL::getMouseClick(this, glm::vec2(x, y));
      computePosition();
    }
  }*/
}

GLvoid OrbitCamera::mouseMove(GLint mouseButton, GLint x, GLint y) {
  Settings* settings = &Settings::getInstance();
  Window* window = opengl->getWindow();

  if (mouseButton & MK_RBUTTON) {
    int button = (int)(log((float)(mouseButton & MK_RBUTTON)) / log(2.0f));
    
    azimuth = azimuthOld + (opengl->getWindow()->getMouse(button).getX() - x) / 200.0f;

    GLfloat altitude_temp = altitudeOld + (y -opengl->getWindow()->getMouse(button).getY()) / 200.0f;
    if (altitude_temp > -(Consts::PI / 2.0) && altitude_temp < Consts::PI / 2.0) {
      altitude = altitude_temp;
    }

    computePosition();
  }
  else if (mouseButton & MK_LBUTTON) {
    glm::vec2 center(settings->getWidth() / 2, settings->getHeight() / 2);
    glm::vec3 delta((lastMouseMovePos.x - x) * (freeCamera ? 0.1f : distance) * 0.005f, (y - lastMouseMovePos.y) * (freeCamera ? 0.1f : distance) * 0.005f, 0);

    glm::vec3 forwardVec = glm::normalize(lookAt - position);
    glm::vec3 sideVec = glm::normalize(glm::cross(forwardVec, up));
    glm::vec3 heightVec = glm::normalize(glm::cross(sideVec, forwardVec));

    glm::mat3 basisMat(sideVec, heightVec, forwardVec);

    lookAt += basisMat * delta;
    position += basisMat * delta;
  }

  wasMouseMove = true;
  lastMouseMovePos = glm::vec2(x, y);
}

GLvoid OrbitCamera::mouseScroll(GLfloat delta) {
  if (!freeCamera) {
    if (distance + delta > 0.0) {
      distance += delta; 
      computePosition();
    }
  }
  else {
    glm::vec3 moveTo = glm::vec3(-delta) * glm::normalize(lookAt - position);
    lookAt += moveTo;
    position += moveTo;
  }
}


FreeCamera::FreeCamera(OpenGL* ogl, glm::vec3& position, glm::vec3& lookDirection, glm::vec3& up, GLfloat moveSpeed, GLfloat sensitivity):
  Camera(ogl, position, lookDirection, up),
  screenPosition(0.0, 0.0),
  moveSpeed(moveSpeed),
  sensitivity(sensitivity) {

  lookAt = position + lookDirection;
}

GLvoid FreeCamera::rotateCamera() {
  Settings* settings = &Settings::getInstance();

  Window* window = opengl->getWindow();

  int x, y;

  if (window->getCursorPosWindow(&x, &y) && window->getActive()) {
    glm::vec2 cursor(x, y);

    glm::vec2 cursorCenter(Settings::getInstance().getWidth() / 2, Settings::getInstance().getHeight() / 2);

    if (cursor.x != cursorCenter.x || cursor.y != cursorCenter.y) {
      cursor.x = sensitivity * (cursor.x - cursorCenter.x) + cursorCenter.x;
      cursor.y = sensitivity * (cursor.y - cursorCenter.y) + cursorCenter.y;

      glPushMatrix();

      glViewport(0, 0, settings->getWidth(), settings->getHeight());	

	    glMatrixMode(GL_PROJECTION);	
	    glLoadIdentity();		
      gluPerspective(60.0f, (GLfloat)settings->getWidth() / (GLfloat)settings->getHeight(), 1.0f, 100.0f);

      glMatrixMode(GL_MODELVIEW);				
	    glLoadIdentity();		

      Camera::useCamera();

      GLint viewport[4];
      GLdouble modelview[16];
      GLdouble projection[16];
      GLdouble posX, posY, posZ;
 
      glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
      glGetDoublev(GL_PROJECTION_MATRIX, projection);
      glGetIntegerv(GL_VIEWPORT, viewport);
 
      gluUnProject((GLdouble)cursor.x, (GLdouble)viewport[3] - (GLdouble)cursor.y, 1.0, modelview, projection, viewport, &posX, &posY, &posZ);
      
      glPopMatrix();

      GLfloat dot = glm::dot(glm::normalize(up), glm::normalize(glm::vec3(glm::vec3(posX, posY, posZ) - position)));
      if (dot > -0.999 && dot < 0.999) {
        lookAt.x = posX;
        lookAt.y = posY;
        lookAt.z = posZ;

        lookAt = position + glm::normalize(lookAt - position);
      }

      window->centerCursor();

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(60.0f, (GLfloat)settings->getWidth() / (GLfloat)settings->getHeight(), opengl->getZNear(), opengl->getZFar());
    }
  }
}

GLvoid FreeCamera::moveCamera() {
  Window* window = opengl->getWindow();

  glm::vec3 moveTo(0.0, 0.0, 0.0);
  glm::vec3 moveSide(0.0, 0.0, 0.0);
  glm::vec3 moveHeight(0.0, 0.0, 0.0);
  
  int i = 0, j = 0, k = 0;

  if (window->getKeyAsync(VK_UP) && !window->getKeyAsync(VK_DOWN)) i = 1;
  else if (window->getKeyAsync(VK_DOWN) && !window->getKeyAsync(VK_UP)) i = -1;
  
  if (window->getKeyAsync(VK_LEFT) && !window->getKeyAsync(VK_RIGHT)) j = 1;
  else if (window->getKeyAsync(VK_RIGHT) && !window->getKeyAsync(VK_LEFT)) j = -1;

  if (window->getKeyAsync(VK_SHIFT) && !window->getKeyAsync(VK_CONTROL)) k = 1;
  else if (window->getKeyAsync(VK_CONTROL) && !window->getKeyAsync(VK_SHIFT)) k = -1;

  if (i != 0) {
    moveTo = glm::vec3(i * moveSpeed) * glm::normalize(lookAt - position);
  }
  if (j != 0) {
    glm::vec3 forward = glm::normalize(lookAt - position);
    moveSide = glm::vec3(j * moveSpeed) * glm::normalize(glm::cross(up, forward));
  }
  if (k != 0) {
    glm::vec3 forward = glm::normalize(lookAt - position);
    glm::vec3 side = glm::normalize(glm::cross(forward, up));
    moveHeight = glm::vec3(k * moveSpeed) * glm::normalize(glm::cross(side, forward));
  }

  lookAt += moveTo + moveSide + moveHeight;
  position += moveTo + moveSide + moveHeight;
}

GLvoid FreeCamera::useCamera() {
  moveCamera();
  rotateCamera();
  Camera::useCamera();
}

GLvoid FreeCamera::keyDown(GLint key) {

}

GLvoid FreeCamera::keyUp(GLint key) {

}

GLvoid FreeCamera::mouseDown(GLint mouseButton, GLint x, GLint y) { 
  
}

GLvoid FreeCamera::mouseUp(GLint mouseButton, GLint x, GLint y) {

}

GLvoid FreeCamera::mouseMove(GLint mouseButton, GLint x, GLint y) { 

}

GLvoid FreeCamera::mouseScroll(GLfloat delta) { 
  
}