#include "../main.h"

OpenGL::OpenGL(Window* wd): 
  window(wd),
  quadric(NULL),
  lastTicks(0), 
  fps(0), 
  frames(0),
  zNear(0.0001f),
  zFar(100000.0f) {
    
  shadersManager = new ShadersManager();
  camerasManager = new CamerasManager(this);
  lightsManager = new LightsManager();

  material = new Material();
}

OpenGL::~OpenGL() {
  delete shadersManager;
  delete camerasManager;
  delete lightsManager;

  delete material;
}

GLvoid OpenGL::createFont2D(GLuint &font2D, std::string fontName, GLint fontSize, GLboolean fontBold) {
  std::ostringstream ss;
  ss << "Building 2D font...";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  if (fontSize > 0) {
    font2D = glGenLists(128);    

    HFONT font = CreateFont(-fontSize,
      0,// Sirka
      0,// Uhol escapement
      0,// Uhol orientacie
      fontBold ? FW_BOLD : 0,// Tucnost
      FALSE,// Kurziva
      FALSE,// Podciarknutie
      FALSE,// Preskrtnutie
      ANSI_CHARSET,// Znakova sada
      OUT_TT_PRECIS,// Presnost vystupu (TrueType)
      CLIP_DEFAULT_PRECIS,// Presnost orezania
      ANTIALIASED_QUALITY,// Vystupna kvalita
      FF_DONTCARE | DEFAULT_PITCH,// Rodina a pitch
      fontName.c_str());// Nazov fontu
  
    HFONT oldFont = (HFONT)SelectObject(window->getHDC(), font);

    if (!wglUseFontBitmaps(window->getHDC(), 32, 128, font2D)) {
      wglUseFontBitmaps(window->getHDC(), 32, 128, font2D);
    }

    DeleteObject(oldFont);
    DeleteObject(font);
  }
}

GLvoid OpenGL::deleteFont2D(GLuint &font2D) {
  std::ostringstream ss;
  ss << "Deleting 2D font...";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  if (font2D != 0) {
    glDeleteLists(font2D, 96);
  }
}

GLvoid OpenGL::countFPS() {
  if (GetTickCount() - lastTicks >= 1000) {
    lastTicks = GetTickCount();
    fps = frames;
    frames = 0;
  }

  frames++;
}

GLvoid OpenGL::glPrint2D(GLuint font, const GLchar *fmt, ...) {
  GLchar text[256];
  va_list ap;

  if (fmt == NULL) {
    return;
  }

  va_start(ap, fmt);
    vsprintf_s(text, fmt, ap);
  va_end(ap);

  glPushAttrib(GL_LIST_BIT);
  glListBase(font - 32);
  
  glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
  glPopAttrib();
}

GLvoid OpenGL::glPrint2D(GLuint font, const std::string& str) {
  glPushAttrib(GL_LIST_BIT);
  glListBase(font - 32);

  glCallLists(str.length(), GL_UNSIGNED_BYTE, str.c_str());
  glPopAttrib();
}

GLboolean OpenGL::initOpenGL() {
  std::ostringstream ss;
  ss << "Loading scene";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  if (!camerasManager->initCameras()) {
    return false;
  }

  if (!lightsManager->initLights()) {
    return false;
  }

  if (!shadersManager->createShaders()) {
    return false;
  }

  if (Settings::getInstance().getMSAA() > 0) {
    glEnable(GL_MULTISAMPLE);
    glEnable(GL_SAMPLE_SHADING);
    glMinSampleShading(2.0f);
    glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
  }
  else {
    glDisable(GL_MULTISAMPLE);
    glDisable(GL_SAMPLE_SHADING);
    glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
  }

  glEnable(GL_DEPTH_TEST);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glEnable(GL_NORMALIZE);

  glEnable(GL_COLOR_MATERIAL);

  glColorMaterial(GL_FRONT, GL_DIFFUSE);

  glClearDepth(1.0f);
  glDepthFunc(GL_LEQUAL);
  glDepthMask(GL_TRUE);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
  glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
  glLineWidth(1.0f);

  glEnable(GL_POINT_SMOOTH);
  glPointSize(1.0f);  

  quadric = gluNewQuadric();

  glm::vec4 materialAmbient(0.2, 0.3, 0.9, 1.0);
  glm::vec4 materialDiffuse(0.2, 0.3, 0.9, 1.0);
  glm::vec4 materialSpecular(1.0, 1.0, 1.0, 1.0);
  glm::vec4 materialEmission(0.0, 0.0, 0.0, 1.0);

  material->loadMaterial(materialAmbient, materialDiffuse, materialSpecular, materialEmission, 64.0, GL_FILL, true);

  return true;
}

GLvoid OpenGL::freeOpenGL() {
  std::ostringstream ss;
  ss << "Deleting scene...";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  if (quadric != NULL) {
    gluDeleteQuadric(quadric);
  }

  camerasManager->freeCameras();
  lightsManager->freeLights();
  shadersManager->deleteShaders();
}

GLvoid OpenGL::createFonts() {
  std::ostringstream ss;
  ss << "Creating fonts...";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  createFont2D(fpsFont2D, "Trebuchet MS", 34);
  createFont2D(infoFont2D, "Tahoma", 12);
}

GLvoid OpenGL::deleteFonts() {
  std::ostringstream ss;
  ss << "Deleting fonts...";
  Log::getInstance()->writeLine(GLDebug, ss.str());

  deleteFont2D(fpsFont2D);
  deleteFont2D(infoFont2D);
}

GLvoid OpenGL::updateSize() {
  GLsizei width = Settings::getInstance().getWidth();
  GLsizei height = Settings::getInstance().getHeight();

  if (width > 0 && height > 0) {
    std::ostringstream ss;
    ss << "Resizing viewport to " << width << "x" << height << "...";
    Log::getInstance()->writeLine(GLDebug, ss.str());
  }
}

GLvoid OpenGL::renderScene(Scene* scene) {
  RenderOptions* renderOptions = &RenderOptions::getInstance();
  Settings* settings = &Settings::getInstance();

  // Clear frame buffer
  glClearColor(0.15f, 0.15f, 0.15f, 1.0f);
  glClearDepth(1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set projection and modelview matrix
  glViewport(0, 0, settings->getWidth(), settings->getHeight());

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0f, (GLfloat)settings->getWidth() / (GLfloat)settings->getHeight(), zNear, zFar);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Render scene
  camerasManager->getCurrentCamera()->useCamera();

  renderOptions->setRenderOptions(shadersManager->getProgram("Simple"), settings->getShowWireframe());
  renderOptions->applyRenderOptions();

  Object* aquarium = scene->getAquarium();
  std::vector<Object>& objects = scene->getObjects();

  const char* data = scene->getFluidParticlesData();
  const char* dataOther = scene->getFluidParticlesDataOther();
  size_t dataSize = scene->getFluidParticlesDataSize();
  size_t dataOtherSize = scene->getFluidParticlesDataOtherSize();
  int particlesCount = scene->getFluidParticlesCount();

  size_t typeSize = dataSize / particlesCount;
  size_t typeSizeOther = dataOtherSize / particlesCount;

  const char* solidData = scene->getSolidParticlesData();
  size_t solidDataSize = scene->getSolidParticlesDataSize();
  int solidParticlesCount = scene->getSolidParticlesCount();

  size_t solidTypeSize = solidDataSize / solidParticlesCount;

  glColor3f(0.8f, 0.8f, 0.8f);

  if (Settings::getInstance().getMSAA() > 0) {
    glDisable(GL_MULTISAMPLE);
    glDisable(GL_SAMPLE_SHADING);
    glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
  }

  aquarium->drawEdges();
  for (int i = 0; i < objects.size(); i++) {
    objects[i].drawEdges();
  }

  if (Settings::getInstance().getMSAA() > 0) {
    glEnable(GL_MULTISAMPLE);
    glEnable(GL_SAMPLE_SHADING);
    glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
  }

  if (!settings->getSpheres()) {
    if (Settings::getInstance().getMSAA() > 0) {
      glDisable(GL_MULTISAMPLE);
      glDisable(GL_SAMPLE_SHADING);
      glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
    }

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    if (settings->getShowSolidParticles()) {
      if (solidData != NULL) {
        glColor3f(0.9f, 0.9f, 0.9f);

        glPointSize(settings->getPointSize());
        glBegin(GL_POINTS);
        for (int i = 0; i < solidParticlesCount; i++) {
          glVertex3f((*(Vector<3, float>*)(void*)(solidData + typeSize * i)).x, (*(Vector<3, float>*)(void*)(solidData + typeSize * i)).y, (*(Vector<3, float>*)(void*)(solidData + typeSize * i)).z);
        }
        glEnd();
        glPointSize(1.0f);
      }
    }

    if (data != NULL) {
      glPointSize(settings->getPointSize());
      glBegin(GL_POINTS);
        if (settings->getFluidParticlesVisualization() == 1) {
          Vector<3, float>* vd;
          Vector<3, float>* vdo;

          float other2;

          for (int i = 0; i < particlesCount; i++) {
            vd = (Vector<3, float>*)(data + typeSize * i);
            vdo = (Vector<3, float>*)(dataOther + typeSize * i);

            other2 = sqrt(length2(vdo->x, vdo->y, vdo->z));
            glColor3f(other2, other2, 1.0f - other2);
            glVertex3f(vd->x, vd->y, vd->z);
          }
        }
        else if (settings->getFluidParticlesVisualization() == 2) {
          Vector<3, float>* vd;
          Vector<3, float>* vdo;

          float other2;
          float x, y, z;

          for (int i = 0; i < particlesCount; i++) {
            vd = (Vector<3, float>*)(data + typeSize * i);
            vdo = (Vector<3, float>*)(dataOther + typeSize * i);

            x = abs(vdo->x); y = abs(vdo->y); z = abs(vdo->z);
            other2 = sqrt(length2(x, y, z));
            if (other2 == 0) { other2 = 1.0f;  }
            glColor3f(((float)(x / other2) + 1.0f) / 2.0f, ((float)(y / other2) + 1.0f) / 2.0f, ((float)(z / other2) + 1.0f) / 2.0f);
            glVertex3f(vd->x, vd->y, vd->z);
          }
        }
        else {
          Vector<3, float>* vd;

          glColor3f(0.0f, 0.0f, 1.0f);
          
          for (int i = 0; i < particlesCount; i++) {
            vd = (Vector<3, float>*)(data + typeSize * i);

            glVertex3f(vd->x, vd->y, vd->z);
          }
        }
      glEnd();
      glPointSize(1.0f);
    }

    glDisable(GL_BLEND);

    if (Settings::getInstance().getMSAA() > 0) {
      glEnable(GL_MULTISAMPLE);
      glEnable(GL_SAMPLE_SHADING);
      glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE_ARB);
    }
  }
  else {
    if (!settings->getShowWireframe()) {
      renderOptions->setRenderOptions(shadersManager->getProgram("Lambert"), settings->getShowWireframe());
    }
    else {
      renderOptions->setRenderOptions(shadersManager->getProgram("Simple"), settings->getShowWireframe());
      glDisable(GL_LINE_SMOOTH);
    }
    renderOptions->applyRenderOptions();
    lightsManager->applyEnabledLights(camerasManager->getCurrentCamera());

    if (settings->getShowSolidParticles()) {
      if (solidData != NULL) {
        material->setDiffuse(0.9f, 0.9f, 0.9f, 1.0f);
        material->setAmbient(0.9f, 0.9f, 0.9f, 1.0f);
        material->applyMaterial();

        for (int i = 0; i < solidParticlesCount; i++) {
          glPushMatrix();
          glTranslatef((*(Vector<3, float>*)(void*)(solidData + typeSize * i)).x, (*(Vector<3, float>*)(void*)(solidData + typeSize * i)).y, (*(Vector<3, float>*)(void*)(solidData + typeSize * i)).z);
          gluSphere(quadric, 0.002f, 10, 10);
          glPopMatrix();
        }
      }
    }

    if (data != NULL) {
      if (settings->getFluidParticlesVisualization() == 1) {
        Vector<3, float>* vd;
        Vector<3, float>* vdo;

        float other2;

        for (int i = 0; i < particlesCount; i++) {
          vd = (Vector<3, float>*)(data + typeSize * i);
          vdo = (Vector<3, float>*)(dataOther + typeSize * i);

          other2 = sqrt(length2(vdo->x, vdo->y, vdo->z));
          material->setDiffuse(std::min(other2, 1.0f), std::min(other2, 1.0f), std::max(1.0f - other2, 0.0f), 1.0f);
          material->setAmbient(std::min(other2, 1.0f), std::min(other2, 1.0f), std::max(1.0f - other2, 0.0f), 1.0f);
          material->applyMaterial();

          glPushMatrix();
          glTranslatef(vd->x, vd->y, vd->z);
          gluSphere(quadric, 0.002f, 10, 10);
          glPopMatrix();
        }
      }
      else if (settings->getFluidParticlesVisualization() == 2) {
        Vector<3, float>* vd;
        Vector<3, float>* vdo;

        float other2;
        float x, y, z;

        for (int i = 0; i < particlesCount; i++) {
          vd = (Vector<3, float>*)(data + typeSize * i);
          vdo = (Vector<3, float>*)(dataOther + typeSize * i);
          x = abs(vdo->x); y = abs(vdo->y); z = abs(vdo->z);

          other2 = sqrt(length2(x, y, z));
          if (other2 == 0) { other2 = 1.0f; }
          material->setDiffuse(((float)(x / other2) + 1.0f) / 2.0f, ((float)(y / other2) + 1.0f) / 2.0f, ((float)(z / other2) + 1.0f) / 2.0f, 1.0f);
          material->setAmbient(((float)(x / other2) + 1.0f) / 2.0f, ((float)(y / other2) + 1.0f) / 2.0f, ((float)(z / other2) + 1.0f) / 2.0f, 1.0f);
          material->applyMaterial();
          
          glPushMatrix();
          glTranslatef(vd->x, vd->y, vd->z);
          gluSphere(quadric, 0.002f, 10, 10);
          glPopMatrix();
        }
      }
      else {
        Vector<3, float>* vd;

        material->setDiffuse(0.2f, 0.3f, 0.9f, 1.0f);
        material->setAmbient(0.2f, 0.3f, 0.9f, 1.0f);
        material->applyMaterial();

        for (int i = 0; i < particlesCount; i++) {
          vd = (Vector<3, float>*)(data + typeSize * i);

          glPushMatrix();
          glTranslatef(vd->x, vd->y, vd->z);
          gluSphere(quadric, 0.002f, 10, 10);
          glPopMatrix();
        }
      }

      material->cancelMaterial();
    }

    lightsManager->cancelEnabledLights();
    renderOptions->cancelRenderOptions();
    glEnable(GL_LINE_SMOOTH);
  }

  if (settings->getShowLights()) {
    lightsManager->renderEnabledLights(camerasManager->getCurrentCamera());
  }

  if (settings->getShowAxes()) {
    renderAxes(0.1f);
  }

  // Finish rendering
  glFinish();
}

GLvoid OpenGL::renderAxes(GLfloat scale) {
  glPushMatrix();

  glScalef(scale, scale, scale);

  glBegin(GL_LINES);
    glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glVertex3f(0.0f, 0.0f, 0.0f);
    glVertex3f(1.0f, 0.0f, 0.0f);

    glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
    glVertex3f(0.0f, 0.0f, 0.0f);
    glVertex3f(0.0f, 1.0f, 0.0f);

    glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
    glVertex3f(0.0f, 0.0f, 0.0f);
    glVertex3f(0.0f, 0.0f, 1.0f);
  glEnd();

  glPopMatrix();
}

GLvoid OpenGL::renderFPS() {
  Settings* settings = &Settings::getInstance();

  // Compute current FPS
  countFPS();  

  // Set projection and modelview matrix
  glMatrixMode(GL_PROJECTION);						
	glLoadIdentity();		
  gluOrtho2D(0, settings->getWidth(), settings->getHeight(), 0);
	glMatrixMode(GL_MODELVIEW);		
  glLoadIdentity();

  glUseProgramObjectARB(0);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
    
  GLint rx = 11, ry = 23;

  rx = settings->getWidth() - 80, ry = 50;
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos3i(rx, ry, 0);
  glPrint2D(fpsFont2D, "%d", fps);
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx-2, ry-2, 0);
  glPrint2D(fpsFont2D, "%d", fps);

  glEnable(GL_DEPTH_TEST);

  // Finish rendering
  glFinish();
}

GLvoid OpenGL::renderInfo(Scene* scene) {
  Settings* settings = &Settings::getInstance();

  // Set projection and modelview matrix
  glMatrixMode(GL_PROJECTION);						
	glLoadIdentity();		
  gluOrtho2D(0, settings->getWidth(), settings->getHeight(), 0);
	glMatrixMode(GL_MODELVIEW);		
  glLoadIdentity();  

  glUseProgramObjectARB(0);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
    
  std::string value;
  std::ostringstream info;
  GLint rx = 10, ry = 20;

  if (settings->getCamera() == 0) {
    value = "orbit";
  }
  else if (settings->getCamera() == 1) {
    value = "orbit free";
  }
  else if (settings->getCamera() == 2) {
    value = "free";
  }

  info << "Camera (C): " << value;
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx-1, ry-1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Spheres (S): " << ((settings->getSpheres()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getSpheres(), 0.8f + 0.2f * settings->getSpheres(), 0.8f + 0.2f * settings->getSpheres());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Show axes (H): " << ((settings->getShowAxes()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getShowAxes(), 0.8f + 0.2f * settings->getShowAxes(), 0.8f + 0.2f * settings->getShowAxes());
  glRasterPos3i(rx-1, ry-1, 1);
  glPrint2D(infoFont2D, info.str());
  
  ry += 20;
  info.str("");  info << "Show wireframe (W): " << ((settings->getShowWireframe()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getShowWireframe(), 0.8f + 0.2f * settings->getShowWireframe(), 0.8f + 0.2f * settings->getShowWireframe());
  glRasterPos3i(rx-1, ry-1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Show lights (L): " << ((settings->getShowLights()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getShowLights(), 0.8f + 0.2f * settings->getShowLights(), 0.8f + 0.2f * settings->getShowLights());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());


  ry += 40;
  info.str("");  info << "Simulation time: " << std::fixed << std::setprecision(3) << settings->getSimulationTime() << " s";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getSimulation(), 0.8f + 0.2f * settings->getSimulation(), 0.8f + 0.2f * settings->getSimulation());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Device (D): " << ((settings->getCurrentDevice() == 0) ? "cpu" : "gpu");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Device local size: " << settings->getCurrentDeviceLocalSize();
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  if (settings->getSimulationMethod() == 0) {
    value = "WCSPH";
  }
  else if (settings->getSimulationMethod() == 1) {
    value = "PCISPH";
  }
  else if (settings->getSimulationMethod() == 2) {
    value = "IISPH";
  }

  ry += 20;
  info.str("");  info << "Simulation method (M): " << value;
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Compute grid every frame method (G): " << (settings->getComputeGridEveryFrame() ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getComputeGridEveryFrame(), 0.8f + 0.2f * settings->getComputeGridEveryFrame(), 0.8f + 0.2f * settings->getComputeGridEveryFrame());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Scene name: " << scene->getName();
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Fluid particles: " << scene->getFluidParticlesCount();
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Solid particles: " << scene->getSolidParticlesCount();
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Buffers memory: " << std::fixed << std::setprecision(3) << (settings->getBuffersMemory() / 1048576.0f) << " MB";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  if (settings->getFluidParticlesVisualization() == 0) {
    value = "solid color";
  }
  else if (settings->getFluidParticlesVisualization() == 1) {
    value = "velocities length";
  }
  else if (settings->getFluidParticlesVisualization() == 2) {
    value = "velocities direction";
  }

  ry += 20;
  info.str("");  info << "Fluid particles vis. (P): " << value << " (" << std::fixed << std::setprecision(0) << settings->getPointSize() << ")";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Show solid particles (O): " << ((settings->getShowSolidParticles()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getShowSolidParticles(), 0.8f + 0.2f * settings->getShowSolidParticles(), 0.8f + 0.2f * settings->getShowSolidParticles());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Vortex (V): " << ((settings->getVortex()) ? "on" : "off");
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(0.8f + 0.2f * settings->getVortex(), 0.8f + 0.2f * settings->getVortex(), 0.8f + 0.2f * settings->getVortex());
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 40;
  info.str("");  info << "Merge sort pass 1: " << settings->getMergeSort1Time() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Merge sort pass 2: " << settings->getMergeSort2Time() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Compute grid overall: " << settings->getComputeGridTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Neighbours: " << settings->getNeighboursTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Compute density: " << settings->getComputeDensityTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Compute force: " << settings->getComputeForceTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Integration: " << settings->getIntegrateTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "State managing: " << settings->getStateManagingTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Simulation overall: " << settings->getOverallComputationTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Download: " << settings->getDownloadTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Realloc neighbours buffer: " << settings->getReallocNeighboursBufferTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  ry += 20;
  info.str("");  info << "Rendering: " << settings->getRenderingTime() << " ms";
  glColor3f(0.0f, 0.0f, 0.0f);
  glRasterPos2i(rx, ry);
  glPrint2D(infoFont2D, info.str());
  glColor3f(1.0f, 1.0f, 1.0f);
  glRasterPos3i(rx - 1, ry - 1, 1);
  glPrint2D(infoFont2D, info.str());

  glEnable(GL_DEPTH_TEST);
  
  // Finish rendering
  glFinish();
}

GLvoid OpenGL::renderInitInfo(std::string str) {
  Settings* settings = &Settings::getInstance();

  glClearColor(0.2, 0.2, 0.2, 1.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set projection and modelview matrix
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, settings->getWidth(), settings->getHeight(), 0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glColor3f(1.0f, 1.0f, 1.0f);

  glRasterPos2i(settings->getWidth() / 2 - 70, settings->getHeight() / 2 - 8);
  glPrint2D(infoFont2D, str.c_str());

  glFinish();
  SwapBuffers(window->getHDC());
}

GLvoid OpenGL::render(Scene* scene) {
  Settings* settings = &Settings::getInstance();

  renderScene(scene);

  if (settings->getShowFPS()) {
    renderFPS();	
  }
  if (settings->getShowInfo()) {
    renderInfo(scene);	
  }
}

GLvoid OpenGL::keyDown(GLint key) {
  camerasManager->keyDown(key);
  lightsManager->keyDown(key);
}

GLvoid OpenGL::keyUp(GLint key) {
  camerasManager->keyUp(key);
  lightsManager->keyUp(key);
}

GLvoid OpenGL::mouseDown(GLint mouseButton, GLint x, GLint y) { 
  camerasManager->mouseDown(mouseButton, x, y);
  lightsManager->mouseDown(mouseButton, x, y);
}

GLvoid OpenGL::mouseUp(GLint mouseButton, GLint x, GLint y) { 
  camerasManager->mouseUp(mouseButton, x, y);
  lightsManager->mouseUp(mouseButton, x, y);
}

GLvoid OpenGL::mouseMove(GLint mouseButton, GLint x, GLint y) { 
  camerasManager->mouseMove(mouseButton, x, y);
  lightsManager->mouseMove(mouseButton, x, y);
}

GLvoid OpenGL::mouseScroll(GLfloat delta) { 
  camerasManager->mouseScroll(delta);
  lightsManager->mouseScroll(delta);
}

GLhandleARB OpenGL::getProgram(std::string programName) {
  return shadersManager->getProgram(programName);
}

GLvoid OpenGL::applyBillboard(Camera* camera, glm::vec3 position) {
  glm::vec3 objToCameraProj = camera->getPosition() - position;
  objToCameraProj.y = 0.0;
  objToCameraProj = glm::normalize(objToCameraProj);
  glm::vec3 objLookAt(0.0, 0.0, 1.0);

  glm::vec3 cross = glm::cross(objLookAt, objToCameraProj);
  GLdouble angleCos = OpenGL::angleCos(objLookAt, objToCameraProj);

  glRotated(acos(angleCos) * 180.0 / Consts::PI, cross.x, cross.y, cross.z);

  glm::vec3 objToCamera = glm::normalize(camera->getPosition() - position);

  angleCos = OpenGL::angleCos(objToCameraProj, objToCamera);

  if (objToCamera.y < 0.0) {
    glRotated(acos(angleCos) * 180.0 / Consts::PI, 1.0, 0.0, 0.0);
  }
  else {
    glRotated(acos(angleCos) * 180.0 / Consts::PI, -1.0, 0.0, 0.0);
  }
}

glm::vec3 OpenGL::getMouseClick(Camera* camera, glm::vec2 cursor, GLboolean withDepth) {
  Settings* settings = &Settings::getInstance();

  glPushMatrix();

  glViewport(0, 0, settings->getWidth(), settings->getHeight());	

	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();		
  gluPerspective(60.0f, (GLfloat)settings->getWidth() / (GLfloat)settings->getHeight(), 1.0f, 100000.0f);

  glMatrixMode(GL_MODELVIEW);				
	glLoadIdentity();		

  camera->useCamera();

  GLint viewport[4];
  GLdouble modelview[16];
  GLdouble projection[16];
  GLdouble posX, posY, posZ;
 
  glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
  glGetDoublev(GL_PROJECTION_MATRIX, projection);
  glGetIntegerv(GL_VIEWPORT, viewport);

  GLfloat depth = 1.0;
  if (withDepth) {
    glReadPixels((GLdouble)cursor.x, (GLdouble)viewport[3] - (GLdouble)cursor.y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
  }
 
  gluUnProject((GLdouble)cursor.x, (GLdouble)viewport[3] - (GLdouble)cursor.y, (GLdouble)depth, modelview, projection, viewport, &posX, &posY, &posZ);

  glPopMatrix();

  glm::vec3 result(posX, posY, posZ);
  return result;
}

GLdouble OpenGL::angleCos(glm::vec3 a, glm::vec3 b) {
  GLdouble da = sqrt(pow(a[0], 2) + pow(a[1], 2) + pow(a[2], 2));
  GLdouble db = sqrt(pow(b[0], 2) + pow(b[1], 2) + pow(b[2], 2));

  return (GLdouble)(a[0] * b[0] + a[1] * b[1] + a[2] * b[2]) / (da * db);
}

float OpenGL::length2(float vecX, float vecY, float vecZ) {
  return vecX * vecX + vecY * vecY + vecZ * vecZ;
}