#include "../main.h"

ShadersManager::ShadersManager() {

}

const char* ShadersManager::loadTextFile(const char* fname) {
	FILE *file = fopen(fname, "rb");

	if (!file) {
    std::ostringstream ss;
    ss << "File " << fname << " is not opened!";
    Log::getInstance()->writeLine(LogType::GLError, ss.str());
		return 0;
	}

	fseek(file, 0, SEEK_END);
	unsigned int nSiz = ftell(file);
	fseek(file, 0, SEEK_SET);
	char* str = new char[nSiz+1];
	fread(str, nSiz, 1, file);
	str[nSiz] = 0;
	fclose(file);

	return str;
}

void ShadersManager::printInfoLog(std::string programName, const GLhandleARB obj) {
  int infoLogLength = 0;
	int charsWritten = 0;
	char *infoLog;

	glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infoLogLength);

	if (infoLogLength > 1) {
    std::ostringstream ss;

		infoLog = (char*)malloc(infoLogLength);

		if (!infoLog) {
      ss << programName.c_str() << " allocation memory failed.";
      Log::getInstance()->writeLine(LogType::GLError, ss.str());
			return;
		}

		glGetInfoLogARB(obj, infoLogLength, &charsWritten, infoLog);

    std::string infoLogStr(infoLog);
    infoLogStr = infoLogStr.substr(0, infoLogStr.find_last_of('\n')-1);

    ss.str("");
    ss << programName.c_str() << " shaderlog (" << infoLogLength << "): " << infoLogStr;
    Log::getInstance()->writeLine(LogType::GLDebug, ss.str());

		free(infoLog);
	}
}

GLhandleARB ShadersManager::getShader(std::string programName, const char* fname, const GLenum shaderType) {
	const char* shaderSrc = loadTextFile(fname);
	
  if (!shaderSrc) {
    return 0;
  }

	GLhandleARB shaderObj = glCreateShaderObjectARB(shaderType);
	glShaderSourceARB(shaderObj, 1, &shaderSrc, NULL);
	glCompileShaderARB(shaderObj);
	printInfoLog(programName, shaderObj);
  delete [] shaderSrc;
	return shaderObj;
}

bool ShadersManager::createProgram(std::string programName, std::string vertexShader, std::string fragmentShader) {
  std::ostringstream ss;
  ss << "Creating program " << programName.c_str() << "...";
  Log::getInstance()->writeLine(LogType::GLDebug, ss.str());

  std::map<std::string, GLhandleARB>::iterator programsIt = programs.find(programName);

  if (programsIt != programs.end()) {
    ss.str("");
    ss << "Creating program " << programName.c_str() << " failed: program with this name already exists.";
    Log::getInstance()->writeLine(LogType::GLError, ss.str());
    return false;
  }

	//create and compile shader OBJECT
  GLhandleARB shaderObjVert = getShader(programName, vertexShader.c_str(), GL_VERTEX_SHADER_ARB);
  GLhandleARB shaderObjFrag = getShader(programName, fragmentShader.c_str(), GL_FRAGMENT_SHADER_ARB);

  if (shaderObjVert == 0 || shaderObjFrag == 0) {
    std::cerr << "Loading shader source code failed!" << std::endl;
    return false;
  }

	//create shader program
	GLhandleARB GLSLprogram = glCreateProgramObjectARB();

	//attach shader object to shader program
  glAttachObjectARB(GLSLprogram, shaderObjVert);
	glAttachObjectARB(GLSLprogram, shaderObjFrag);
  
	//link shader program
	glLinkProgramARB(GLSLprogram);

	//print shader log (errors)
	printInfoLog(programName, GLSLprogram);

  GLint linked;
  glGetProgramiv(GLSLprogram, GL_LINK_STATUS, &linked);
  if (!linked) {
    return false;
  }  
  
  glUseProgramObjectARB(GLSLprogram);
  glUseProgramObjectARB(0);

  vertexShaders.insert(std::pair<std::string, GLhandleARB>(programName, shaderObjVert));
  fragmentShaders.insert(std::pair<std::string, GLhandleARB>(programName, shaderObjFrag));
  programs.insert(std::pair<std::string, GLhandleARB>(programName, GLSLprogram));

  return true;
}

bool ShadersManager::deleteProgram(std::string programName) {
  std::map<std::string, GLhandleARB>::iterator programsIt = programs.find(programName);

  if (programsIt != programs.end()) {
    std::map<std::string, GLhandleARB>::iterator vertexShadersIt = vertexShaders.find(programName);
    if (vertexShadersIt != vertexShaders.end()) {
      glDeleteShader(vertexShadersIt->second);
      vertexShaders.erase(vertexShadersIt);
    }

    std::map<std::string, GLhandleARB>::iterator fragmentShadersIt = fragmentShaders.find(programName);
    if (fragmentShadersIt != fragmentShaders.end()) {
      glDeleteShader(fragmentShadersIt->second);
      fragmentShaders.erase(fragmentShadersIt);
    }

    glDeleteProgram(programsIt->second);
    programs.erase(programsIt);

    return true;
  }

  return false;
}

GLhandleARB ShadersManager::getProgram(std::string programName) {
  std::map<std::string, GLhandleARB>::iterator programsIt = programs.find(programName);

  if (programsIt != programs.end()) {
    return programsIt->second;
  }
  else {
    return 0;
  }
}

bool ShadersManager::createShaders() {
  std::ostringstream ss;
  ss << "Creating shaders...";
  Log::getInstance()->writeLine(LogType::GLDebug, ss.str());

  if (!createProgram("Simple", "shaders/simple.vert", "shaders/simple.frag")) {
    return false;
  }
  if (!createProgram("Lambert", "shaders/lambert.vert", "shaders/lambert.frag")) {
    return false;
  }
  if (!createProgram("Blinn-Phong", "shaders/blinn_phong.vert", "shaders/blinn_phong.frag")) {
    return false;
  }

  return true;
}

void ShadersManager::deleteShaders() {
  std::ostringstream ss;
  ss << "Deleting shaders...";
  Log::getInstance()->writeLine(LogType::GLDebug, ss.str());

  deleteProgram("Simple");
  deleteProgram("Lambert");
  deleteProgram("Blinn-Phong");
}