#include "../main.h"

Object::~Object() {
  clear();
}

void Object::clear() {
  vertices.clear();
  edges.clear();
  triangles.clear();
}

void Object::computeEdges() {
  edges.clear();

  std::map<int, std::map<int, bool>> pointEdgeMap;

  bool created;
  Vector<3, int> t;
  Vector<2, int> edge;

  for (int i = 0; i < triangles.size(); i++) {
    t = triangles[i];

    created = pointEdgeMap[t.x][t.y];
    if (!created) {
      pointEdgeMap[t.x][t.y] = 1;
      pointEdgeMap[t.y][t.x] = 1;
      edge.x = t.x;
      edge.y = t.y;
      edges.push_back(edge);
    }

    created = pointEdgeMap[t.y][t.z];
    if (!created) {
      pointEdgeMap[t.y][t.z] = 1;
      pointEdgeMap[t.z][t.y] = 1;
      edge.x = t.y;
      edge.y = t.z;
      edges.push_back(edge);
    }

    created = pointEdgeMap[t.z][t.x];
    if (!created) {
      pointEdgeMap[t.z][t.x] = 1;
      pointEdgeMap[t.x][t.z] = 1;
      edge.x = t.z;
      edge.y = t.x;
      edges.push_back(edge);
    }
  }
}

void Object::fillTriangleWithParticles(Vector<3, float> v0, Vector<3, float> v1, Vector<3, float> v2, float distance, std::vector<Vector<3, float>>& particles) {
  Vector<3, float> longest = v2 - v0;
  Vector<3, float> shortest = v1 - v0;
  Vector<3, float> middle = v2 - v1;
  Vector<3, float> inside = shortest.cross(longest.cross(shortest));

  inside = inside.normalize();

  float height = inside.dot(longest);
  int steps = height / distance;

  Vector<3, float> vec1 = longest / steps;
  Vector<3, float> vec2 = middle / steps;
  Vector<3, float> pos1;
  Vector<3, float> pos2;
  Vector<3, float> vec;
  float sweepSteps;

  Vector<3, float> p;

  for (int i = 1; i < steps; i++) {
    pos1 = v0 + vec1*i;
    pos2 = v1 + vec2*i;
    vec = pos2 - pos1;
    sweepSteps = vec.length() / distance;
    vec /= sweepSteps;

    for (int j = 1; j < sweepSteps; j++) {
      p = pos1 + vec * j;
      particles.push_back(p);
    }
  }
}

bool Object::loadFromFile(std::string filePath) {
  Log* log = Log::getInstance();

  clear();

  std::ifstream inStream;
  char line[256];

  inStream.open(filePath);
  if (!inStream.is_open()) {
    log->writeLine(SPHDebug, "Error while loading object from file '" + filePath + "': file not found!");
    return false;
  }

  int lineNumber = 0;
  while (!inStream.eof()) {
    inStream.getline(line, 256);

    //read vertex (v x y z)
    if (line[0] == 'v') {
      Vector<3, float> v;
      int start = 2;
      int i = 2;
      std::stringstream ss;

      while (line[i] != ' ') {
        ss.put(line[i]);
        i++;
      }
      ss >> v.x;
      ss.clear();
      i++;

      while (line[i] != ' ') {
        ss.put(line[i]);
        i++;
      }
      ss >> v.y;
      ss.clear();
      i++;

      while (line[i] != '\0') {
        ss.put(line[i]);
        i++;
      }
      ss >> v.z;

      vertices.push_back(v);
    }

    //read face (f i1 i2 i3   or   f i1 i2 i3 i4)
    if (line[0] == 'f') {
      Vector<3, float> t;
      int start = 2;
      int i = 2;
      std::stringstream ss;
      int i1, i2, i3, i4;

      while (line[i] != ' ') {
        ss.put(line[i]);
        i++;
      }
      ss >> i1;
      ss.clear();
      i++;

      while (line[i] != ' ') {
        ss.put(line[i]);
        i++;
      }
      ss >> i2;
      ss.clear();
      i++;

      while (line[i] != '\0' && line[i] != ' ') {
        ss.put(line[i]);
        i++;
      }
      ss >> i3;

      //obj file is index from 1 while vector is indexing from 0
      i1--;
      i2--;
      i3--;

      t.x = i1;
      t.y = i2;
      t.z = i3;
      triangles.push_back(t);

      // read fourth index if needed
      if (line[i] == ' ') {
        ss.clear();
        i++;
        while (line[i] != '\0') {
          ss.put(line[i]);
          i++;
        }
        ss >> i4;
        i4--;

        t.x = i1;
        t.y = i3;
        t.z = i4;
        triangles.push_back(t);
      }
    }

    lineNumber++;
  }

  computeEdges();

  return true;
}

std::vector<Vector<3, float>> Object::generateSurfaceParticles(float distance) {
  std::vector<Vector<3, float>> particles;

  //seed particles on vertex positions
  Vector<3, float> p;
  Vector<3, float> v;
  for (int i = 0; i < vertices.size(); i++) {
    v = vertices[i];
    p.x = v.x;
    p.y = v.y;
    p.z = v.z;
    particles.push_back(p);
  }

  //seed particles along edges
  Vector<2, int> e;
  Vector<3, float> v0, v1;
  for (int i = 0; i < edges.size(); i++) {
    e = edges[i];
    Vector<3, float> v0 = vertices[e.x];
    Vector<3, float> v1 = vertices[e.y];
    float dist = (v1 - v0).length();

    if (dist > distance) {
      int steps = dist / distance;
      Vector<3, float> vec = (v1 - v0) / steps;
      for (int s = 1; s < steps; s++) {
        p = v0 + s*vec;
        particles.push_back(p);
      }
    }
  }

  //seed particles inside triangles
  Vector<3, int> t;
  Vector<3, float> v2;
  for (int i = 0; i < triangles.size(); i++) {
    t = triangles[i];
    v0 = vertices[t.x];
    v1 = vertices[t.y];
    v2 = vertices[t.z];
    float a = (v1 - v0).length();
    float b = (v2 - v1).length();
    float c = (v0 - v2).length();

    if (a < b) {
      if (a < c) {
        if (c < b) {
          fillTriangleWithParticles(v1, v0, v2, distance, particles);
        }
        else {
          fillTriangleWithParticles(v0, v1, v2, distance, particles);
        }
      }
      else {
        fillTriangleWithParticles(v0, v2, v1, distance, particles);
      }
    }
    else {
      if (b < c) {
        if (c < a) {
          fillTriangleWithParticles(v1, v2, v0, distance, particles);
        }
        else {
          fillTriangleWithParticles(v2, v1, v0, distance, particles);
        }
      }
      else {
        fillTriangleWithParticles(v0, v2, v1, distance, particles);
      }
    }
  }

  return particles;
}

void Object::scale(float scaleFactor) {
  for (int i = 0; i < vertices.size(); i++) {
    vertices[i] *= scaleFactor;
  }
}

void Object::scale(Vector<3, float> scaleVector) {
  for (int i = 0; i < vertices.size(); i++) {
    vertices[i] *= scaleVector;
  }
}

void Object::translate(Vector<3, float> translationVector) {
  for (int i = 0; i < vertices.size(); i++) {
    vertices[i] += translationVector;
  }
}

void Object::drawEdges() {
  Vector<2, int>* e;
  Vector<3, float>* v0;
  Vector<3, float>* v1;

  glBegin(GL_LINES);
  for (int i = 0; i < edges.size(); i++) {
    e = &edges[i];
    v0 = &vertices[e->x];
    v1 = &vertices[e->y];
    glVertex3f(v0->x, v0->y, v0->z);
    glVertex3f(v1->x, v1->y, v1->z);
  }
  glEnd();
}