#include "sph.h"

SPHsim::SPHsim(Window* wd):
  window(wd), 
  reduceLocalSize(256),
  mergeSortLocalSize(256) {

  Log::getInstance()->supressLogType(LogType::CLDebug);

  scenesPointers.push_back(&SPHsim::setSceneDamBreak);
  scenesPointers.push_back(&SPHsim::setSceneDamBreakObstacle);
  scenesPointers.push_back(&SPHsim::setSceneDamBreakTwoFluids);
  scenesPointers.push_back(&SPHsim::setSceneOcean);
  scenesPointers.push_back(&SPHsim::setSceneOceanBig);
  scenesPointers.push_back(&SPHsim::setSceneDrop);
}

SPHsim::~SPHsim() {

}

float SPHsim::distance2(const Vector<3, float>& a, const Vector<3, float>& b) {
  float abX = b.x - a.x;
  float abY = b.y - a.y;
  float abZ = b.z - a.z;
  return abX * abX + abY * abY + abZ * abZ;
}

int SPHsim::getBuffersSize() {
  Settings* settings = &Settings::getInstance();

  int size = 0;

  size += density->getMemorySize();
  size += densityFluctuation->getMemorySize();
  size += isoDensity->getMemorySize();
  size += pressure->getMemorySize();
  
  size += pos->getMemorySize();
  size += posTmp->getMemorySize();
  size += vel->getMemorySize();
  size += velTmp->getMemorySize();
  size += velXSPH->getMemorySize();
  size += acc->getMemorySize();
  size += accTmp->getMemorySize();
  size += vorticity->getMemorySize();

  size += gridCellsFluidA->getMemorySize();
  size += gridCellsFluidB->getMemorySize();
  size += gridCellIndicesFluid->getMemorySize();

  size += gridCellsSolidA->getMemorySize();
  size += gridCellsSolidB->getMemorySize();
  size += gridCellIndicesSolid->getMemorySize();

  size += neighboursFluidFluid->getMemorySize();
  size += neighboursFluidFluidCount->getMemorySize();
  size += neighboursFluidSolid->getMemorySize();
  size += neighboursFluidSolidCount->getMemorySize();
  size += neighboursSolidSolid->getMemorySize();
  size += neighboursSolidSolidCount->getMemorySize();

  size += solidPos->getMemorySize();
  size += solidPosTmp->getMemorySize();

  // PCISPH || IISPH
  if (settings->getSimulationMethod() == 1 || settings->getSimulationMethod() == 2) {
    size += pressureForce->getMemorySize();
  }

  // IISPH
  if (settings->getSimulationMethod() == 2) {
    size += velocityPrediction->getMemorySize();
    size += densityPrediction->getMemorySize();
    size += d_ii->getMemorySize();
    size += a_ii->getMemorySize();
    size += d_ijp_j->getMemorySize();
    size += sumA->getMemorySize();
  }

  return size;
}

std::vector<Vector<3, float>> SPHsim::createFluidBox(Vector<3, float> minCell, Vector<3, float> size, Vector<3, float> cut, float jitter, float radius) {
  srand(1);

  std::vector<Vector<3, float>> result;

  Vector<3, int> c = (2 * minCell + size);
  Vector<3, float> center = 0.5f * Vector<3, float>(c.x, c.y, c.z);
  float radius2 = radius*radius;

  for (int x = 0; x < size.x; x++) {
    for (int y = 0; y < size.y; y++) {
      for (int z = 0; z < size.z; z++) {
        if (x >= size.x - cut.x && y >= size.y - cut.y && z >= size.z - cut.z) continue;

        Vector<3, float> position = Vector<3, float>(minCell.x + x, minCell.y + y, minCell.z + z);
        if (distance2(center, position) > radius2) continue;

        Vector<3, float> finalPosition = distance0Host * position;

        if (jitter != 0) {
          finalPosition += Vector<3, float>(jitter*((float)rand() / RAND_MAX), jitter*((float)rand() / RAND_MAX), jitter*((float)rand() / RAND_MAX));
        }
        result.push_back(finalPosition);
      }
    }
  }

  return result;
}

void SPHsim::initializeParticles(std::vector<Vector<3, float>>& fluidParticles, std::vector<Vector<3, float>>& solidParticles) {
  Settings* settings = &Settings::getInstance();

  for (unsigned int i = 0; i < scene.getFluidParticlesCount(); i++) {
    density->setScalar(i, density0Host);
    densityFluctuation->setScalar(i, 0);
    isoDensity->setScalar(i, density0Host);
    pressure->setScalar(i, 0);

    pos->setVector(i, fluidParticles[i]);
    posTmp->setVector(i, fluidParticles[i]);
    vel->setVector(i, Vector<3, float>(0, 0, 0));
    velTmp->setVector(i, Vector<3, float>(0, 0, 0));
    velXSPH->setVector(i, Vector<3, float>(0, 0, 0));
    acc->setVector(i, Vector<3, float>(0, 0, 0));
    accTmp->setVector(i, Vector<3, float>(0, 0, 0));
    vorticity->setVector(i, Vector<3, float>(0, 0, 0));
  }

  // PCISPH || IISPH
  if (settings->getSimulationMethod() == 1 || settings->getSimulationMethod() == 2) {
    for (unsigned int i = 0; i < scene.getFluidParticlesCount(); i++) {
      pressureForce->setVector(i, Vector<3, float>(0, 0, 0));
    }
  }

  // IISPH
  if (settings->getSimulationMethod() == 2) {
    for (unsigned int i = 0; i < scene.getFluidParticlesCount(); i++) {
      velocityPrediction->setVector(i, Vector<3, float>(0, 0, 0));
      densityPrediction->setScalar(i, 0);
      d_ii->setVector(i, Vector<3, float>(0, 0, 0));
      a_ii->setScalar(i, 0);
      d_ijp_j->setVector(i, Vector<3, float>(0, 0, 0));
      sumA->setScalar(i, 0);
    }
  }

  for (unsigned int i = 0; i < scene.getSolidParticlesCount(); i++) {
    solidPos->setVector(i, solidParticles[i]);
    solidPosTmp->setVector(i, Vector<2, double>(0, 0));
  }

  for (unsigned int i = 0; i < mergeSortFluidGlobalSize; i++) {
    gridCellsFluidA->setVector(i, Vector<2, double>(0, 0));
    gridCellsFluidB->setVector(i, Vector<2, double>(0, 0));
  }

  for (unsigned int i = 0; i < mergeSortSolidGlobalSize; i++) {
    gridCellsSolidA->setVector(i, Vector<2, double>(0, 0));
    gridCellsSolidB->setVector(i, Vector<2, double>(0, 0));
  }

  density->upload();
  pressure->upload();
  pos->upload();
  posTmp->upload();
  vel->upload();
  velTmp->upload();
  velXSPH->upload();
  acc->upload();
  accTmp->upload();
  vorticity->upload();

  solidPos->upload();

  gridCellsFluidA->upload(); gridCellsFluidB->upload();
  gridCellsSolidA->upload(); gridCellsSolidB->upload();

  // PCISPH || IISPH
  if (settings->getSimulationMethod() == 1 || settings->getSimulationMethod() == 2) {
    pressureForce->upload();
  }

  // IISPH
  if (settings->getSimulationMethod() == 2) {
    velocityPrediction->upload();
    densityPrediction->upload();
    d_ii->upload();
    a_ii->upload();
    d_ijp_j->upload();
    sumA->upload();
  }

  scene.setFluidParticlesData(pos->getData(), pos->getMemorySize());
  scene.setFluidParticlesDataOther(vel->getData(), vel->getMemorySize());
  scene.setSolidParticlesData(solidPos->getData(), solidPos->getMemorySize());
}

void SPHsim::precomputeDelta() {
  float beta = mass0Host * mass0Host * 2.0 / (density0Host * density0Host);
  std::vector<Vector<3, float>> tempParticles;

  for (int i = 0; i < 3; i++) {
    for (int j = 0; j < 3; j++) {
      for (int k = 0; k < 3; k++) {
        tempParticles.push_back(Vector<3, float>(0.99*i*distance0Host, 0.99*j*distance0Host, 0.99*k*distance0Host));
      }
    }
  }

  Vector<3, float> kernelGrad(0, 0, 0);
  float kernelGrad2 = 0;
  Vector<3, float> grad;
  float density = 0;
  float invRadius = 1.0f / radiusHost;
  float invRadius2 = invRadius * invRadius;

  Vector<3, float> p = tempParticles[13];

  for (int i = 0; i < tempParticles.size(); i++) {
    grad = p - tempParticles[i];
    float dist = grad.length();

    if (dist < radiusHost && dist != 0) {
      float hr = 1 - dist * invRadius;
      float t = 1 - dist*dist*invRadius2;
      grad /= dist;
      grad *= hr*hr;
      
      kernelGrad += grad;
      kernelGrad2 += grad.dot(grad);
      density += mass0Host * wDensityHost * t*t*t;
    }
  }

  density += mass0Host * wDensityHost;
  kernelGrad *= dWPressureHost;
  kernelGrad2 *= dWPressureHost * dWPressureHost;

  deltaHost = -1.0f / (beta * (-kernelGrad.dot(kernelGrad) - kernelGrad2));
}

void SPHsim::reallocateNeighboursBuffers(bool neighboursCountComputed) {
  Settings* settings = &Settings::getInstance();

  if (!neighboursCountComputed) {
    if (programComputeNeighboursFluidCount->getIsBuilt()) {
      programComputeNeighboursFluidCount->enqueue();
      programComputeNeighboursFluidCount->finish();
    }

    if (programComputeNeighboursSolidCount->getIsBuilt()) {
      programComputeNeighboursSolidCount->enqueue();
      programComputeNeighboursSolidCount->finish();
    }

    neighboursFluidFluidCountMax->setScalar(0);
    neighboursFluidSolidCountMax->setScalar(0);
    neighboursSolidSolidCountMax->setScalar(0);
  }
  
  int fluidFluidMax = 0, fluidSolidMax = 0, solidSolidMax = 0;
  int fluidFluidMaxLast = 0, fluidSolidMaxLast = 0, solidSolidMaxLast = 0;

  // Fluid-fluid
  if ((!programComputeNeighboursFluidCount->getIsBuilt() || !programReduceMaxFluidFluidCount->getIsBuilt()) && !neighboursCountComputed) {
    fluidFluidMax = 30;
  }
  else {
    reduceLength->setScalar(scene.getFluidParticlesCount());
    programReduceMaxFluidFluidCount->enqueue(reduceLocalSize, reduceLocalSize);
    programReduceMaxFluidFluidCount->finish();

    reduceResultInt->download(true, true);
    fluidFluidMax = (int)reduceResultInt->getScalar(0) * 1.2;

    if (fluidFluidMax > scene.getFluidParticlesCount()) fluidFluidMax = scene.getFluidParticlesCount();
  }

  fluidFluidMaxLast = neighboursFluidFluidCountMax->getScalar();
  if (fluidFluidMaxLast < (fluidFluidMax / 1.2) * 1.1 || (fluidFluidMaxLast / 1.2) > (fluidFluidMax / 1.2) * 1.1) {
    neighboursFluidFluidCountMax->setScalar(fluidFluidMax);
    neighboursFluidFluid->reallocate(fluidFluidMax * scene.getFluidParticlesCount());
  }

  // Fluid-solid
  if ((!programComputeNeighboursFluidCount->getIsBuilt() || !programReduceMaxFluidSolidCount->getIsBuilt()) && !neighboursCountComputed) {
    fluidSolidMax = 30;
  }
  else {
    reduceLength->setScalar(scene.getFluidParticlesCount());
    programReduceMaxFluidSolidCount->enqueue(reduceLocalSize, reduceLocalSize);
    programReduceMaxFluidSolidCount->finish();

    reduceResultInt->download(true, true);
    fluidSolidMax = (int)reduceResultInt->getScalar(0) * 1.2;

    if (fluidSolidMax > scene.getSolidParticlesCount()) fluidSolidMax = scene.getSolidParticlesCount();
  }

  fluidSolidMaxLast = neighboursFluidSolidCountMax->getScalar();
  if (fluidSolidMaxLast < (fluidSolidMax / 1.2) * 1.1 || (fluidSolidMaxLast / 1.2) > (fluidSolidMax / 1.2) * 1.1) {
    neighboursFluidSolidCountMax->setScalar(fluidSolidMax);
    neighboursFluidSolid->reallocate(fluidSolidMax * scene.getFluidParticlesCount());
  }

  // Solid-solid
  if ((!programComputeNeighboursSolidCount->getIsBuilt() || !programReduceMaxSolidSolidCount->getIsBuilt()) && !neighboursCountComputed) {
    solidSolidMax = 30;
  }
  else {
    reduceLength->setScalar(scene.getSolidParticlesCount());
    programReduceMaxSolidSolidCount->enqueue(reduceLocalSize, reduceLocalSize);
    programReduceMaxSolidSolidCount->finish();

    reduceResultInt->download(true, true);
    solidSolidMax = (int)reduceResultInt->getScalar(0) * 1.2;

    if (solidSolidMax > scene.getSolidParticlesCount()) solidSolidMax = scene.getSolidParticlesCount();
  }

  solidSolidMaxLast = neighboursSolidSolidCountMax->getScalar();
  if (solidSolidMaxLast < (solidSolidMax / 1.2) * 1.1 || (solidSolidMaxLast / 1.2) > (solidSolidMax / 1.2) * 1.1) {
    neighboursSolidSolidCountMax->setScalar(solidSolidMax);
    neighboursSolidSolid->reallocate(solidSolidMax * scene.getSolidParticlesCount());
  }

  //std::cout << "fluidFluidMax: " << fluidFluidMax << "  fluidSolidMax: " << fluidSolidMax << "  solidSolidMax: " << solidSolidMax << std::endl;
  settings->setBuffersMemory(getBuffersSize());
}

void SPHsim::computeGridFluid() {
  Settings* settings = &Settings::getInstance();

  programCreateCellsFluid->enqueue();
  programCreateCellsFluid->finish();

  if (mergeSortFluidDummyElements > 0) {
    correctElements->setScalar(scene.getFluidParticlesCount());
    programCorrectCellsFluid->enqueue(mergeSortFluidDummyElements, 1);
    programCorrectCellsFluid->finish();
  }

  double t = clock();
  programMergeSortCellsFluid->enqueue(mergeSortFluidGlobalSize, mergeSortLocalSize);
  programMergeSortCellsFluid->finish();
  settings->setMergeSort1Time(clock() - t);

  int mergeSize = mergeSortLocalSize;
  mergeSortMergeTotalSize->setScalar(mergeSortFluidGlobalSize);

  CLProgram* programMergeSortMerge = (mergeSortFluidMergesCount % 2 == 0) ? programMergeSortMergeCellsFluidA : programMergeSortMergeCellsFluidB;

  t = clock();
  while (mergeSize < mergeSortFluidGlobalSize) {
    mergeSortMergeMergeSize->setScalar(mergeSize);
    mergeSize <<= 1;

    programMergeSortMerge->enqueue((mergeSortFluidGlobalSize / mergeSize) + 1, 1);
    programMergeSortMerge->finish();

    programMergeSortMerge = (programMergeSortMerge == programMergeSortMergeCellsFluidA) ? programMergeSortMergeCellsFluidB : programMergeSortMergeCellsFluidA;
  }
  settings->setMergeSort2Time(clock() - t);

  mergeSortMergeCellsFluidLastCell->download(true, true);
  int gridCellIndicesFluidUsedSize = mergeSortMergeCellsFluidLastCell->getScalar(0) + 1;

  if (((CLVariable*)gridCellIndicesFluid)->getElementCount() < gridCellIndicesFluidUsedSize) {
    gridCellIndicesFluid->reallocate(gridCellIndicesFluidUsedSize);
  }

  programClearCellIndicesFluid->enqueue(gridCellIndicesFluidUsedSize, mergeSortLocalSize);
  programClearCellIndicesFluid->finish();

  dummyElements->setScalar(mergeSortFluidDummyElements);
  programCreateCellIndicesFluid->enqueue(mergeSortFluidGlobalSize, mergeSortLocalSize);
  programCreateCellIndicesFluid->finish();

  /*std::ostringstream ss;
  ss << framesHost;

  if (framesHost > 260) {
    pos->download(true, true);

    std::ofstream file("outFluida" + ss.str() + ".txt");
    for (int i = 0; i < scene.getFluidParticlesCount(); i++) {
      file << pos->getVector(i).x << " " << pos->getVector(i).y << " " << pos->getVector(i).z << std::endl;
    }
    file.close();
  }*/

  programSwapFluid->enqueue(mergeSortFluidGlobalSize, mergeSortLocalSize);
  programSwapFluid->finish();

  pos->copyFrom(posTmp);
  vel->copyFrom(velTmp);

  /*if (framesHost > 260) {
    pos->download(true, true);

    std::ofstream file2("outFluidb" + ss.str() + ".txt");
    for (int i = 0; i < scene.getFluidParticlesCount(); i++) {
      file2 << pos->getVector(i).x << " " << pos->getVector(i).y << " " << pos->getVector(i).z << std::endl;
    }
    file2.close();
  }*/

  /*gridCellIndicesFluid->download(true, true);

  std::ofstream file("outFluid.txt");
  for (int i = 0; i < mergeSortMergeCellsFluidLastCell->getScalar(0) + 1; i++) {
    file << gridCellIndicesFluid->getScalar(i) << std::endl;
  }
  file.close();*/
}

void SPHsim::computeGridSolid() {
  programCreateCellsSolid->enqueue();
  programCreateCellsSolid->finish();

  if (mergeSortSolidDummyElements > 0) {
    correctElements->setScalar(scene.getSolidParticlesCount());
    programCorrectCellsSolid->enqueue(mergeSortSolidDummyElements, 1);
    programCorrectCellsSolid->finish();
  }

  programMergeSortCellsSolid->enqueue(mergeSortSolidGlobalSize, mergeSortLocalSize);
  programMergeSortCellsSolid->finish();

  int mergeSize = mergeSortLocalSize;
  mergeSortMergeTotalSize->setScalar(mergeSortSolidGlobalSize);

  CLProgram* programMergeSortMerge = (mergeSortSolidMergesCount % 2 == 0) ? programMergeSortMergeCellsSolidA : programMergeSortMergeCellsSolidB;

  while (mergeSize < mergeSortSolidGlobalSize) {
    mergeSortMergeMergeSize->setScalar(mergeSize);
    mergeSize <<= 1;

    programMergeSortMerge->enqueue((mergeSortSolidGlobalSize / mergeSize) + 1, 1);
    programMergeSortMerge->finish();

    programMergeSortMerge = (programMergeSortMerge == programMergeSortMergeCellsSolidA) ? programMergeSortMergeCellsSolidB : programMergeSortMergeCellsSolidA;
  }

  mergeSortMergeCellsSolidLastCell->download(true, true);
  int gridCellIndicesSolidUsedSize = mergeSortMergeCellsSolidLastCell->getScalar(0) + 1;

  if (((CLVariable*)gridCellIndicesSolid)->getElementCount() < gridCellIndicesSolidUsedSize) {
    gridCellIndicesSolid->reallocate(gridCellIndicesSolidUsedSize);
  }

  programClearCellIndicesSolid->enqueue(gridCellIndicesSolidUsedSize, mergeSortLocalSize);
  programClearCellIndicesSolid->finish();

  dummyElements->setScalar(mergeSortSolidDummyElements);
  programCreateCellIndicesSolid->enqueue(mergeSortSolidGlobalSize, mergeSortLocalSize);
  programCreateCellIndicesSolid->finish();

  programSwapSolid->enqueue(mergeSortSolidGlobalSize, mergeSortLocalSize);
  programSwapSolid->finish();

  solidPos->copyFrom(solidPosTmp);

  /*gridCellIndicesSolid->download(true, true);

  std::ofstream file("outSolid.txt");
  for (int i = 0; i < gridCellIndicesSolidUsedSize; i++) {
    file << gridCellIndicesSolid->getScalar(i) << std::endl;
  }
  file.close();*/
}

bool SPHsim::createOpenCLObjects(CLLink* link) {
  Settings* settings = &Settings::getInstance();

  bool status = true;

  reduceLocalSize = link->getDevice(0)->getMaxWorkItemSize(0);
  mergeSortLocalSize = link->getDevice(0)->getMaxWorkItemSize(0);

  mergeSortFluidGlobalSize = (scene.getFluidParticlesCount() % mergeSortLocalSize) ? ((int)((scene.getFluidParticlesCount() / mergeSortLocalSize) + 1) * mergeSortLocalSize) : (scene.getFluidParticlesCount());
  mergeSortFluidDummyElements = mergeSortFluidGlobalSize - scene.getFluidParticlesCount();
  mergeSortFluidMergesCount = ((ceil(LOG2(mergeSortFluidGlobalSize)) - (int)LOG2(mergeSortLocalSize)));

  mergeSortSolidGlobalSize = (scene.getSolidParticlesCount() % mergeSortLocalSize) ? ((int)((scene.getSolidParticlesCount() / mergeSortLocalSize) + 1) * mergeSortLocalSize) : (scene.getSolidParticlesCount());
  mergeSortSolidDummyElements = mergeSortSolidGlobalSize - scene.getSolidParticlesCount();
  mergeSortSolidMergesCount = ((ceil(LOG2(mergeSortSolidGlobalSize)) - (int)LOG2(mergeSortLocalSize)));

  // Common
  functionDistance2 = new CLFunction("clsources/functions/distance2.cl", "distance2");
  functionLength2 = new CLFunction("clsources/functions/length2.cl", "length2");
  functionZOrder = new CLFunction("clsources/functions/z_order.cl", "zOrder");

  programReduceMaxFluidFluidCount = new CLProgram("clsources/helpers/reduce_max_int.cl", "reduceMaxInt", link);
  programReduceMaxFluidSolidCount = new CLProgram("clsources/helpers/reduce_max_int.cl", "reduceMaxInt", link);
  programReduceMaxSolidSolidCount = new CLProgram("clsources/helpers/reduce_max_int.cl", "reduceMaxInt", link);
  programComputeNeighboursFluidCount = new CLProgram("clsources/neighbours/compute_neighbours_fluid_count.cl", "computeNeighboursFluidCount", link);
  programComputeNeighboursSolidCount = new CLProgram("clsources/neighbours/compute_neighbours_solid_count.cl", "computeNeighboursSolidCount", link);
  programCreateCellsFluid = new CLProgram("clsources/neighbours/create_cells.cl", "createCells", link);
  programCreateCellsSolid = new CLProgram("clsources/neighbours/create_cells.cl", "createCells", link);
  programCorrectCellsFluid = new CLProgram("clsources/neighbours/correct_cells.cl", "correctCells", link);
  programCorrectCellsSolid = new CLProgram("clsources/neighbours/correct_cells.cl", "correctCells", link);
  programMergeSortCellsFluid = new CLProgram("clsources/neighbours/merge_sort.cl", "mergeSort", link);
  programMergeSortCellsSolid = new CLProgram("clsources/neighbours/merge_sort.cl", "mergeSort", link);
  programMergeSortMergeCellsFluidA = new CLProgram("clsources/neighbours/merge_sort_merge.cl", "mergeSortMerge", link);
  programMergeSortMergeCellsFluidB = new CLProgram("clsources/neighbours/merge_sort_merge.cl", "mergeSortMerge", link);
  programMergeSortMergeCellsSolidA = new CLProgram("clsources/neighbours/merge_sort_merge.cl", "mergeSortMerge", link);
  programMergeSortMergeCellsSolidB = new CLProgram("clsources/neighbours/merge_sort_merge.cl", "mergeSortMerge", link);
  programSwapFluid = new CLProgram("clsources/neighbours/swap_fluid.cl", "swapFluid", link);
  programSwapSolid = new CLProgram("clsources/neighbours/swap_solid.cl", "swapSolid", link);
  programClearCellIndicesFluid = new CLProgram("clsources/neighbours/clear_cell_indices.cl", "clearCellIndices", link);
  programClearCellIndicesSolid = new CLProgram("clsources/neighbours/clear_cell_indices.cl", "clearCellIndices", link);
  programCreateCellIndicesFluid = new CLProgram("clsources/neighbours/create_cell_indices.cl", "createCellIndices", link);
  programCreateCellIndicesSolid = new CLProgram("clsources/neighbours/create_cell_indices.cl", "createCellIndices", link);
  programFindNeighboursFluid = new CLProgram("clsources/neighbours/find_neighbours_fluid.cl", "findNeighboursFluid", link);
  programFindNeighboursSolid = new CLProgram("clsources/neighbours/find_neighbours_solid.cl", "findNeighboursSolid", link);
  programFindNeighboursFluidN2 = new CLProgram("clsources/neighbours/find_neighbours_fluid_n2.cl", "findNeighboursFluidN2", link);
  programFindNeighboursSolidN2 = new CLProgram("clsources/neighbours/find_neighbours_solid_n2.cl", "findNeighboursSolidN2", link);
  programComputeSolidVolume = new CLProgram("clsources/forces/compute_solid_volume.cl", "computeSolidVolume", link);
  programComputeDensity = new CLProgram("clsources/forces/compute_density.cl", "computeDensity", link);
  programNormalizeDensityFirstPass = new CLProgram("clsources/forces/normalize_density_first_pass.cl", "normalizeDensityFirstPass", link);
  programNormalizeDensitySecondPass = new CLProgram("clsources/forces/normalize_density_second_pass.cl", "normalizeDensitySecondPass", link);
  programComputeForce = new CLProgram("clsources/forces/compute_force.cl", "computeForce", link);
  programIntegratePositions = new CLProgram("clsources/integration/integrate_positions.cl", "integratePositions", link);
  programApplyXSPHFirstPass = new CLProgram("clsources/integration/apply_xsph_first_pass.cl", "applyXSPHFirstPass", link);
  programApplyXSPHSecondPass = new CLProgram("clsources/integration/apply_xsph_second_pass.cl", "applyXSPHSecondPass", link);
  programClearState = new CLProgram("clsources/state_managing/clear_state.cl", "clearState", link);
  
  reduceScratchInt = new CLLocalBuffer("scratch", CLVariableDataType::IntType, reduceLocalSize);
  reduceScratchFloat = new CLLocalBuffer("scratch", CLVariableDataType::FloatType, reduceLocalSize);
  mergeSortAux = new CLLocalBuffer("aux", CLVariableDataType::ULong2Type, mergeSortLocalSize);

  reduceResultInt = new CLGlobalBuffer(link, "result", CLVariableDataType::IntType, 1);
  reduceResultFloat = new CLGlobalBuffer(link, "result", CLVariableDataType::FloatType, 1);

  density = new CLGlobalBuffer(link, "density", CLVariableDataType::FloatType, scene.getFluidParticlesCount(), false, true);
  densityFluctuation = new CLGlobalBuffer(link, "densityFluctuation", CLVariableDataType::FloatType, scene.getFluidParticlesCount());
  isoDensity = new CLGlobalBuffer(link, "isoDensity", CLVariableDataType::FloatType, scene.getFluidParticlesCount());
  pressure = new CLGlobalBuffer(link, "pressure", CLVariableDataType::FloatType, scene.getFluidParticlesCount(), false, true);
  vorticity = new CLGlobalBuffer(link, "vorticity", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());

  pos = new CLGlobalBuffer(link, "pos", CLVariableDataType::Float3Type, scene.getFluidParticlesCount(), false, true);
  posTmp = new CLGlobalBuffer(link, "posTmp", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());
  vel = new CLGlobalBuffer(link, "vel", CLVariableDataType::Float3Type, scene.getFluidParticlesCount(), false, true);
  velTmp = new CLGlobalBuffer(link, "velTmp", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());
  velXSPH = new CLGlobalBuffer(link, "velXSPH", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());
  acc = new CLGlobalBuffer(link, "acc", CLVariableDataType::Float3Type, scene.getFluidParticlesCount(), false, true);
  accTmp = new CLGlobalBuffer(link, "accTmp", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());

  gridCellsFluidA = new CLGlobalBuffer(link, "gridCellsFluidA", CLVariableDataType::ULong2Type, mergeSortFluidGlobalSize);
  gridCellsFluidB = new CLGlobalBuffer(link, "gridCellsFluidB", CLVariableDataType::ULong2Type, mergeSortFluidGlobalSize);
  gridCellIndicesFluid = new CLGlobalBuffer(link, "gridCellIndicesFluid", CLVariableDataType::IntType, 1, true);
  mergeSortMergeCellsFluidLastCell = new CLGlobalBuffer(link, "lastCellFluid", CLVariableDataType::ULongType, 1);

  gridCellsSolidA = new CLGlobalBuffer(link, "gridCellsSolidA", CLVariableDataType::ULong2Type, mergeSortSolidGlobalSize);
  gridCellsSolidB = new CLGlobalBuffer(link, "gridCellsSolidB", CLVariableDataType::ULong2Type, mergeSortSolidGlobalSize);
  gridCellIndicesSolid = new CLGlobalBuffer(link, "gridCellIndicesSolid", CLVariableDataType::IntType, 1, true);
  mergeSortMergeCellsSolidLastCell = new CLGlobalBuffer(link, "lastCellSolid", CLVariableDataType::ULongType, 1);
  
  neighboursFluidFluid = new CLGlobalBuffer(link, "neighboursFluidFluid", CLVariableDataType::IntType, 1, true);
  neighboursFluidFluidCount = new CLGlobalBuffer(link, "neighboursFluidFluidCount", CLVariableDataType::IntType, scene.getFluidParticlesCount());
  neighboursFluidSolid = new CLGlobalBuffer(link, "neighboursFluidSolid", CLVariableDataType::IntType, 1, true);
  neighboursFluidSolidCount = new CLGlobalBuffer(link, "neighboursFluidSolidCount", CLVariableDataType::IntType, scene.getFluidParticlesCount());
  neighboursSolidSolid = new CLGlobalBuffer(link, "neighboursSolidSolid", CLVariableDataType::IntType, 1, true);
  neighboursSolidSolidCount = new CLGlobalBuffer(link, "neighboursSolidSolidCount", CLVariableDataType::IntType, scene.getSolidParticlesCount());

  solidPos = new CLGlobalBuffer(link, "solidPos", CLVariableDataType::Float3Type, scene.getSolidParticlesCount(), false, true);
  solidPosTmp = new CLGlobalBuffer(link, "solidPosTmp", CLVariableDataType::Float3Type, scene.getSolidParticlesCount());

  solidVolume = new CLGlobalBuffer(link, "solidVolume", CLVariableDataType::FloatType, scene.getSolidParticlesCount(), false, true);

  dummyElements = new CLKernelArgument("dummyElements", CLVariableDataType::IntType);
  correctElements = new CLKernelArgument("correctElements", CLVariableDataType::IntType);

  reduceLength = new CLKernelArgument("length", CLVariableDataType::IntType);

  mergeSortMergeMergeSize = new CLKernelArgument("mergeSize", CLVariableDataType::IntType);
  mergeSortMergeTotalSize = new CLKernelArgument("totalSize", CLVariableDataType::IntType);

  usePressure = new CLKernelArgument("usePressure", CLVariableDataType::IntType);
  useViscosity = new CLKernelArgument("useViscosity", CLVariableDataType::IntType);
  useVorticity = new CLKernelArgument("useVorticity", CLVariableDataType::IntType);
  useVortex = new CLKernelArgument("useVortex", CLVariableDataType::IntType);
  computePressure = new CLKernelArgument("computePressure", CLVariableDataType::IntType);
  clearPressure = new CLKernelArgument("clearPressure", CLVariableDataType::IntType);
  neighboursFluidFluidCountMax = new CLKernelArgument("neighboursFluidFluidCountMax", CLVariableDataType::IntType);
  neighboursFluidSolidCountMax = new CLKernelArgument("neighboursFluidSolidCountMax", CLVariableDataType::IntType);
  neighboursSolidSolidCountMax = new CLKernelArgument("neighboursSolidSolidCountMax", CLVariableDataType::IntType);

  fluidParticlesCount = new CLConstant("fluidParticlesCount", CLVariableDataType::UIntType, scene.getFluidParticlesCount());
  solidParticlesCount = new CLConstant("solidParticlesCount", CLVariableDataType::UIntType, scene.getSolidParticlesCount());
  deltaTime = new CLConstant("deltaTime", CLVariableDataType::FloatType, deltaTimeHost);
  deltaTime2 = new CLConstant("deltaTime2", CLVariableDataType::FloatType, deltaTimeHost * deltaTimeHost);

  delta = new CLConstant("delta", CLVariableDataType::FloatType, deltaHost);
  omega = new CLConstant("omega", CLVariableDataType::FloatType, omegaHost);
  gravity = new CLConstant("gravity", CLVariableDataType::FloatType, gravityHost);
  radius = new CLConstant("radius", CLVariableDataType::FloatType, radiusHost);
  radius2 = new CLConstant("radius2", CLVariableDataType::FloatType, radiusHost*radiusHost);
  invRadius = new CLConstant("invRadius", CLVariableDataType::FloatType, 1.0f / radiusHost);
  invRadius2 = new CLConstant("invRadius2", CLVariableDataType::FloatType, 1.0f / (radiusHost*radiusHost));
  accelerationLimit = new CLConstant("accelerationLimit", CLVariableDataType::FloatType, accelerationLimitHost);
  accelerationLimit2 = new CLConstant("accelerationLimit2", CLVariableDataType::FloatType, accelerationLimitHost*accelerationLimitHost);
  wDensity = new CLConstant("wDensity", CLVariableDataType::FloatType, wDensityHost);
  dWDensity = new CLConstant("dWDensity", CLVariableDataType::FloatType, dWDensityHost);
  ddWViscosity = new CLConstant("ddWViscosity", CLVariableDataType::FloatType, ddWViscosityHost);
  dWPressure = new CLConstant("dWPressure", CLVariableDataType::FloatType, dWPressureHost);
  density0 = new CLConstant("density0", CLVariableDataType::FloatType, density0Host);
  mass0 = new CLConstant("mass0", CLVariableDataType::FloatType, mass0Host);
  viscosity0 = new CLConstant("viscosity0", CLVariableDataType::FloatType, viscosity0Host);
  solidViscosity0 = new CLConstant("solidViscosity0", CLVariableDataType::FloatType, solidViscosity0Host);
  stiffness = new CLConstant("stiffness", CLVariableDataType::FloatType, stiffnessHost);
  negativePressureFactor = new CLConstant("negativePressureFactor", CLVariableDataType::FloatType, negativePressureFactorHost);
  seventh = new CLConstant("seventh", CLVariableDataType::FloatType, 1.0f / 7.0f);
  boundaryStiffness = new CLConstant("boundaryStiffness", CLVariableDataType::FloatType, boundaryStiffnessHost);
  vorticityFactor = new CLConstant("vorticityFactor", CLVariableDataType::FloatType, vorticityFactorHost);
  surfaceTension = new CLConstant("surfaceTension", CLVariableDataType::FloatType, surfaceTensionHost);
  vortexOrigin = new CLConstant("vortexOrigin", CLVariableDataType::Float3Type, vortexOriginHost);
  vortexAxis = new CLConstant("vortexAxis", CLVariableDataType::Float3Type, vortexAxisHost);
  vortexFactor = new CLConstant("vortexFactor", CLVariableDataType::FloatType, vortexFactorHost);
  
  programReduceMaxFluidFluidCount->addGlobalBuffer(neighboursFluidFluidCount, 0);
  programReduceMaxFluidFluidCount->addLocalBuffer(reduceScratchInt, 1);
  programReduceMaxFluidFluidCount->addKernelArgument(reduceLength, 2);
  programReduceMaxFluidFluidCount->addGlobalBuffer(reduceResultInt, 3);
  status &= programReduceMaxFluidFluidCount->build();

  programReduceMaxFluidSolidCount->addGlobalBuffer(neighboursFluidSolidCount, 0);
  programReduceMaxFluidSolidCount->addLocalBuffer(reduceScratchInt, 1);
  programReduceMaxFluidSolidCount->addKernelArgument(reduceLength, 2);
  programReduceMaxFluidSolidCount->addGlobalBuffer(reduceResultInt, 3);
  status &= programReduceMaxFluidSolidCount->build();

  programReduceMaxSolidSolidCount->addGlobalBuffer(neighboursSolidSolidCount, 0);
  programReduceMaxSolidSolidCount->addLocalBuffer(reduceScratchInt, 1);
  programReduceMaxSolidSolidCount->addKernelArgument(reduceLength, 2);
  programReduceMaxSolidSolidCount->addGlobalBuffer(reduceResultInt, 3);
  status &= programReduceMaxSolidSolidCount->build();

  programComputeNeighboursFluidCount->addGlobalBuffer(pos, 0);
  programComputeNeighboursFluidCount->addGlobalBuffer(gridCellsFluidA, 1);
  programComputeNeighboursFluidCount->addGlobalBuffer(gridCellIndicesFluid, 2);
  programComputeNeighboursFluidCount->addGlobalBuffer(mergeSortMergeCellsFluidLastCell, 3);
  programComputeNeighboursFluidCount->addGlobalBuffer(neighboursFluidFluidCount, 4);
  programComputeNeighboursFluidCount->addGlobalBuffer(solidPos, 5);
  programComputeNeighboursFluidCount->addGlobalBuffer(gridCellsSolidA, 6);
  programComputeNeighboursFluidCount->addGlobalBuffer(gridCellIndicesSolid, 7);
  programComputeNeighboursFluidCount->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 8);
  programComputeNeighboursFluidCount->addGlobalBuffer(neighboursFluidSolidCount, 9);
  programComputeNeighboursFluidCount->addConstant(radius2);
  programComputeNeighboursFluidCount->addConstant(invRadius);
  programComputeNeighboursFluidCount->addFunction(functionZOrder);
  programComputeNeighboursFluidCount->addFunction(functionDistance2);
  status &= programComputeNeighboursFluidCount->build();

  programComputeNeighboursSolidCount->addGlobalBuffer(solidPos, 0);
  programComputeNeighboursSolidCount->addGlobalBuffer(gridCellsSolidA, 1);
  programComputeNeighboursSolidCount->addGlobalBuffer(gridCellIndicesSolid, 2);
  programComputeNeighboursSolidCount->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 3);
  programComputeNeighboursSolidCount->addGlobalBuffer(neighboursSolidSolidCount, 4);
  programComputeNeighboursSolidCount->addConstant(radius2);
  programComputeNeighboursSolidCount->addConstant(invRadius);
  programComputeNeighboursSolidCount->addFunction(functionZOrder);
  programComputeNeighboursSolidCount->addFunction(functionDistance2);
  status &= programComputeNeighboursSolidCount->build();

  programCreateCellsFluid->addGlobalBuffer(pos, 0);
  if (mergeSortFluidMergesCount % 2 == 0) {
    programCreateCellsFluid->addGlobalBuffer(gridCellsFluidA, 1);
  }
  else {
    programCreateCellsFluid->addGlobalBuffer(gridCellsFluidB, 1);
  }
  programCreateCellsFluid->addConstant(invRadius);
  programCreateCellsFluid->addFunction(functionZOrder);
  status &= programCreateCellsFluid->build();

  programCreateCellsSolid->addGlobalBuffer(solidPos, 0);
  if (mergeSortSolidMergesCount % 2 == 0) {
    programCreateCellsSolid->addGlobalBuffer(gridCellsSolidA, 1);
  }
  else {
    programCreateCellsSolid->addGlobalBuffer(gridCellsSolidB, 1);
  }
  programCreateCellsSolid->addConstant(invRadius);
  programCreateCellsSolid->addFunction(functionZOrder);
  status &= programCreateCellsSolid->build();

  programCorrectCellsFluid->addGlobalBuffer(gridCellsFluidA, 0);
  programCorrectCellsFluid->addGlobalBuffer(gridCellsFluidB, 1);
  programCorrectCellsFluid->addKernelArgument(correctElements, 2);
  status &= programCorrectCellsFluid->build();

  programCorrectCellsSolid->addGlobalBuffer(gridCellsSolidA, 0);
  programCorrectCellsSolid->addGlobalBuffer(gridCellsSolidB, 1);
  programCorrectCellsSolid->addKernelArgument(correctElements, 2);
  status &= programCorrectCellsSolid->build();

  if (mergeSortFluidMergesCount % 2 == 0) {
    programMergeSortCellsFluid->addGlobalBuffer(gridCellsFluidA, 0);
  }
  else {
    programMergeSortCellsFluid->addGlobalBuffer(gridCellsFluidB, 0);
  }
  programMergeSortCellsFluid->addLocalBuffer(mergeSortAux, 1);
  status &= programMergeSortCellsFluid->build();

  if (mergeSortSolidMergesCount % 2 == 0) {
    programMergeSortCellsSolid->addGlobalBuffer(gridCellsSolidA, 0);
  }
  else {
    programMergeSortCellsSolid->addGlobalBuffer(gridCellsSolidB, 0);
  }
  programMergeSortCellsSolid->addLocalBuffer(mergeSortAux, 1);
  status &= programMergeSortCellsSolid->build();

  programMergeSortMergeCellsFluidA->addGlobalBuffer(gridCellsFluidA, 0);
  programMergeSortMergeCellsFluidA->addGlobalBuffer(gridCellsFluidB, 1);
  programMergeSortMergeCellsFluidA->addKernelArgument(mergeSortMergeMergeSize, 2);
  programMergeSortMergeCellsFluidA->addKernelArgument(mergeSortMergeTotalSize, 3);
  programMergeSortMergeCellsFluidA->addGlobalBuffer(mergeSortMergeCellsFluidLastCell, 4);
  status &= programMergeSortMergeCellsFluidA->build();

  programMergeSortMergeCellsFluidB->addGlobalBuffer(gridCellsFluidB, 0);
  programMergeSortMergeCellsFluidB->addGlobalBuffer(gridCellsFluidA, 1);
  programMergeSortMergeCellsFluidB->addKernelArgument(mergeSortMergeMergeSize, 2);
  programMergeSortMergeCellsFluidB->addKernelArgument(mergeSortMergeTotalSize, 3);
  programMergeSortMergeCellsFluidB->addGlobalBuffer(mergeSortMergeCellsFluidLastCell, 4);
  status &= programMergeSortMergeCellsFluidB->build();

  programMergeSortMergeCellsSolidA->addGlobalBuffer(gridCellsSolidA, 0);
  programMergeSortMergeCellsSolidA->addGlobalBuffer(gridCellsSolidB, 1);
  programMergeSortMergeCellsSolidA->addKernelArgument(mergeSortMergeMergeSize, 2);
  programMergeSortMergeCellsSolidA->addKernelArgument(mergeSortMergeTotalSize, 3);
  programMergeSortMergeCellsSolidA->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 4);
  status &= programMergeSortMergeCellsSolidA->build();

  programMergeSortMergeCellsSolidB->addGlobalBuffer(gridCellsSolidB, 0);
  programMergeSortMergeCellsSolidB->addGlobalBuffer(gridCellsSolidA, 1);
  programMergeSortMergeCellsSolidB->addKernelArgument(mergeSortMergeMergeSize, 2);
  programMergeSortMergeCellsSolidB->addKernelArgument(mergeSortMergeTotalSize, 3);
  programMergeSortMergeCellsSolidB->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 4);
  status &= programMergeSortMergeCellsSolidB->build();

  programSwapFluid->addGlobalBuffer(pos, 0);
  programSwapFluid->addGlobalBuffer(posTmp, 1);
  programSwapFluid->addGlobalBuffer(vel, 2);
  programSwapFluid->addGlobalBuffer(velTmp, 3);
  programSwapFluid->addGlobalBuffer(gridCellsFluidA, 4);
  programSwapFluid->addKernelArgument(dummyElements, 5);
  status &= programSwapFluid->build();

  programSwapSolid->addGlobalBuffer(solidPos, 0);
  programSwapSolid->addGlobalBuffer(solidPosTmp, 1);
  programSwapSolid->addGlobalBuffer(gridCellsSolidA, 2);
  programSwapSolid->addKernelArgument(dummyElements, 3);
  status &= programSwapSolid->build();

  programClearCellIndicesFluid->addGlobalBuffer(gridCellIndicesFluid, 0);
  status &= programClearCellIndicesFluid->build();

  programClearCellIndicesSolid->addGlobalBuffer(gridCellIndicesSolid, 0);
  status &= programClearCellIndicesSolid->build();

  programCreateCellIndicesFluid->addGlobalBuffer(gridCellsFluidA, 0);
  programCreateCellIndicesFluid->addGlobalBuffer(gridCellIndicesFluid, 1);
  programCreateCellIndicesFluid->addKernelArgument(dummyElements, 2);
  status &= programCreateCellIndicesFluid->build();

  programCreateCellIndicesSolid->addGlobalBuffer(gridCellsSolidA, 0);
  programCreateCellIndicesSolid->addGlobalBuffer(gridCellIndicesSolid, 1);
  programCreateCellIndicesSolid->addKernelArgument(dummyElements, 2);
  status &= programCreateCellIndicesSolid->build();

  programFindNeighboursFluid->addGlobalBuffer(pos, 0);
  programFindNeighboursFluid->addGlobalBuffer(gridCellsFluidA, 1);
  programFindNeighboursFluid->addGlobalBuffer(gridCellIndicesFluid, 2);
  programFindNeighboursFluid->addGlobalBuffer(mergeSortMergeCellsFluidLastCell, 3);
  programFindNeighboursFluid->addGlobalBuffer(neighboursFluidFluid, 4);
  programFindNeighboursFluid->addGlobalBuffer(neighboursFluidFluidCount, 5);
  programFindNeighboursFluid->addKernelArgument(neighboursFluidFluidCountMax, 6);
  programFindNeighboursFluid->addGlobalBuffer(solidPos, 7);
  programFindNeighboursFluid->addGlobalBuffer(gridCellsSolidA, 8);
  programFindNeighboursFluid->addGlobalBuffer(gridCellIndicesSolid, 9);
  programFindNeighboursFluid->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 10);
  programFindNeighboursFluid->addGlobalBuffer(neighboursFluidSolid, 11);
  programFindNeighboursFluid->addGlobalBuffer(neighboursFluidSolidCount, 12);
  programFindNeighboursFluid->addKernelArgument(neighboursFluidSolidCountMax, 13);
  programFindNeighboursFluid->addConstant(radius2);
  programFindNeighboursFluid->addConstant(invRadius);
  programFindNeighboursFluid->addFunction(functionZOrder);
  programFindNeighboursFluid->addFunction(functionDistance2);
  status &= programFindNeighboursFluid->build();

  programFindNeighboursSolid->addGlobalBuffer(solidPos, 0);
  programFindNeighboursSolid->addGlobalBuffer(gridCellsSolidA, 1);
  programFindNeighboursSolid->addGlobalBuffer(gridCellIndicesSolid, 2);
  programFindNeighboursSolid->addGlobalBuffer(mergeSortMergeCellsSolidLastCell, 3);
  programFindNeighboursSolid->addGlobalBuffer(neighboursSolidSolid, 4);
  programFindNeighboursSolid->addGlobalBuffer(neighboursSolidSolidCount, 5);
  programFindNeighboursSolid->addKernelArgument(neighboursSolidSolidCountMax, 6);
  programFindNeighboursSolid->addConstant(radius2);
  programFindNeighboursSolid->addConstant(invRadius);
  programFindNeighboursSolid->addFunction(functionZOrder);
  programFindNeighboursSolid->addFunction(functionDistance2);
  status &= programFindNeighboursSolid->build();

  programFindNeighboursFluidN2->addGlobalBuffer(pos, 0);
  programFindNeighboursFluidN2->addGlobalBuffer(neighboursFluidFluid, 1);
  programFindNeighboursFluidN2->addGlobalBuffer(neighboursFluidFluidCount, 2);
  programFindNeighboursFluidN2->addKernelArgument(neighboursFluidFluidCountMax, 3);
  programFindNeighboursFluidN2->addGlobalBuffer(solidPos, 4);
  programFindNeighboursFluidN2->addGlobalBuffer(neighboursFluidSolid, 5);
  programFindNeighboursFluidN2->addGlobalBuffer(neighboursFluidSolidCount, 6);
  programFindNeighboursFluidN2->addKernelArgument(neighboursFluidSolidCountMax, 7);
  programFindNeighboursFluidN2->addConstant(fluidParticlesCount);
  programFindNeighboursFluidN2->addConstant(solidParticlesCount);
  programFindNeighboursFluidN2->addConstant(radius2);
  programFindNeighboursFluidN2->addFunction(functionDistance2);
  status &= programFindNeighboursFluidN2->build();

  programFindNeighboursSolidN2->addGlobalBuffer(solidPos, 0);
  programFindNeighboursSolidN2->addGlobalBuffer(neighboursSolidSolid, 1);
  programFindNeighboursSolidN2->addGlobalBuffer(neighboursSolidSolidCount, 2);
  programFindNeighboursSolidN2->addKernelArgument(neighboursSolidSolidCountMax, 3);
  programFindNeighboursSolidN2->addConstant(solidParticlesCount);
  programFindNeighboursSolidN2->addConstant(radius2);
  programFindNeighboursSolidN2->addFunction(functionDistance2);
  status &= programFindNeighboursSolidN2->build();

  programComputeSolidVolume->addGlobalBuffer(solidPos, 0);
  programComputeSolidVolume->addGlobalBuffer(solidVolume, 1);
  programComputeSolidVolume->addGlobalBuffer(neighboursSolidSolid, 2);
  programComputeSolidVolume->addKernelArgument(neighboursSolidSolidCountMax, 3);
  programComputeSolidVolume->addConstant(invRadius2);
  programComputeSolidVolume->addConstant(wDensity);
  programComputeSolidVolume->addConstant(density0);
  programComputeSolidVolume->addFunction(functionLength2);
  status &= programComputeSolidVolume->build();

  programComputeDensity->addGlobalBuffer(pos, 0);
  programComputeDensity->addGlobalBuffer(vel, 1);
  programComputeDensity->addGlobalBuffer(vorticity, 2);
  programComputeDensity->addGlobalBuffer(density, 3);
  programComputeDensity->addKernelArgument(useVorticity, 4);
  programComputeDensity->addGlobalBuffer(neighboursFluidFluid, 5);
  programComputeDensity->addKernelArgument(neighboursFluidFluidCountMax, 6);
  programComputeDensity->addGlobalBuffer(solidPos, 7);
  programComputeDensity->addGlobalBuffer(solidVolume, 8);
  programComputeDensity->addGlobalBuffer(neighboursFluidSolid, 9);
  programComputeDensity->addKernelArgument(neighboursFluidSolidCountMax, 10);
  programComputeDensity->addConstant(invRadius2);
  programComputeDensity->addConstant(wDensity);
  programComputeDensity->addConstant(dWDensity);
  programComputeDensity->addConstant(density0);
  programComputeDensity->addConstant(mass0);
  programComputeDensity->addConstant(stiffness);
  programComputeDensity->addConstant(negativePressureFactor);
  programComputeDensity->addFunction(functionLength2);
  status &= programComputeDensity->build();

  programNormalizeDensityFirstPass->addGlobalBuffer(pos, 0);
  programNormalizeDensityFirstPass->addGlobalBuffer(density, 1);
  programNormalizeDensityFirstPass->addGlobalBuffer(isoDensity, 2);
  programNormalizeDensityFirstPass->addGlobalBuffer(neighboursFluidFluid, 3);
  programNormalizeDensityFirstPass->addKernelArgument(neighboursFluidFluidCountMax, 4);
  programNormalizeDensityFirstPass->addGlobalBuffer(solidPos, 5);
  programNormalizeDensityFirstPass->addGlobalBuffer(solidVolume, 6);
  programNormalizeDensityFirstPass->addGlobalBuffer(neighboursFluidSolid, 7);
  programNormalizeDensityFirstPass->addKernelArgument(neighboursFluidSolidCountMax, 8);
  programNormalizeDensityFirstPass->addConstant(invRadius2);
  programNormalizeDensityFirstPass->addConstant(density0);
  programNormalizeDensityFirstPass->addConstant(mass0);
  programNormalizeDensityFirstPass->addFunction(functionDistance2);
  status &= programNormalizeDensityFirstPass->build();

  programNormalizeDensitySecondPass->addGlobalBuffer(density, 0);
  programNormalizeDensitySecondPass->addGlobalBuffer(isoDensity, 1);
  programNormalizeDensitySecondPass->addGlobalBuffer(densityFluctuation, 2);
  programNormalizeDensitySecondPass->addGlobalBuffer(pressure, 3);
  programNormalizeDensitySecondPass->addKernelArgument(computePressure, 4);
  programNormalizeDensitySecondPass->addConstant(wDensity);
  programNormalizeDensitySecondPass->addConstant(density0);
  programNormalizeDensitySecondPass->addConstant(mass0);
  programNormalizeDensitySecondPass->addConstant(stiffness);
  programNormalizeDensitySecondPass->addConstant(negativePressureFactor);
  programNormalizeDensitySecondPass->addConstant(seventh);
  status &= programNormalizeDensitySecondPass->build();

  programComputeForce->addGlobalBuffer(pos, 0);
  programComputeForce->addGlobalBuffer(vel, 1);
  programComputeForce->addGlobalBuffer(acc, 2);
  programComputeForce->addGlobalBuffer(vorticity, 3);
  programComputeForce->addGlobalBuffer(density, 4);
  programComputeForce->addGlobalBuffer(isoDensity, 5);
  programComputeForce->addGlobalBuffer(pressure, 6);
  programComputeForce->addKernelArgument(usePressure, 7);
  programComputeForce->addKernelArgument(useViscosity, 8);
  programComputeForce->addKernelArgument(useVorticity, 9);
  programComputeForce->addGlobalBuffer(neighboursFluidFluid, 10);
  programComputeForce->addKernelArgument(neighboursFluidFluidCountMax, 11);
  programComputeForce->addGlobalBuffer(solidPos, 12);
  programComputeForce->addGlobalBuffer(solidVolume, 13);
  programComputeForce->addGlobalBuffer(neighboursFluidSolid, 14);
  programComputeForce->addKernelArgument(neighboursFluidSolidCountMax, 15);
  programComputeForce->addConstant(radius);
  programComputeForce->addConstant(radius2);
  programComputeForce->addConstant(invRadius);
  programComputeForce->addConstant(mass0);
  programComputeForce->addConstant(viscosity0);
  programComputeForce->addConstant(solidViscosity0);
  programComputeForce->addConstant(dWDensity);
  programComputeForce->addConstant(ddWViscosity);
  programComputeForce->addConstant(dWPressure);
  programComputeForce->addConstant(surfaceTension);
  programComputeForce->addConstant(vorticityFactor);
  programComputeForce->addConstant(gravity);
  programComputeForce->addFunction(functionLength2);
  status &= programComputeForce->build();

  programIntegratePositions->addGlobalBuffer(pos, 0);
  programIntegratePositions->addGlobalBuffer(vel, 1);
  programIntegratePositions->addConstant(deltaTime);
  status &= programIntegratePositions->build();

  programClearState->addGlobalBuffer(density, 0);
  programClearState->addGlobalBuffer(isoDensity, 1);
  programClearState->addGlobalBuffer(pressure, 2);
  programClearState->addGlobalBuffer(acc, 3);
  programClearState->addGlobalBuffer(vorticity, 4);
  programClearState->addKernelArgument(clearPressure, 5);
  status &= programClearState->build();

  programApplyXSPHFirstPass->addGlobalBuffer(pos, 0);
  programApplyXSPHFirstPass->addGlobalBuffer(vel, 1);
  programApplyXSPHFirstPass->addGlobalBuffer(velXSPH, 2);
  programApplyXSPHFirstPass->addGlobalBuffer(density, 3);
  programApplyXSPHFirstPass->addGlobalBuffer(neighboursFluidFluid, 4);
  programApplyXSPHFirstPass->addKernelArgument(neighboursFluidFluidCountMax, 5);
  programApplyXSPHFirstPass->addConstant(invRadius2);
  programApplyXSPHFirstPass->addFunction(functionLength2);
  status &= programApplyXSPHFirstPass->build();

  programApplyXSPHSecondPass->addGlobalBuffer(vel, 0);
  programApplyXSPHSecondPass->addGlobalBuffer(velXSPH, 1);
  programApplyXSPHSecondPass->addConstant(wDensity);
  programApplyXSPHSecondPass->addConstant(mass0);
  status &= programApplyXSPHSecondPass->build();

  
  // WCSPH
  if (settings->getSimulationMethod() == 0) {
    programIntegrateVelocitiesWCSPH = new CLProgram("clsources/integration/integrate_velocities_wcsph.cl", "integrateVelocitiesWCSPH", link);

    programIntegrateVelocitiesWCSPH->addGlobalBuffer(pos, 0);
    programIntegrateVelocitiesWCSPH->addGlobalBuffer(vel, 1);
    programIntegrateVelocitiesWCSPH->addGlobalBuffer(acc, 2);
    programIntegrateVelocitiesWCSPH->addKernelArgument(useVortex, 3);
    programIntegrateVelocitiesWCSPH->addConstant(deltaTime);
    programIntegrateVelocitiesWCSPH->addConstant(accelerationLimit);
    programIntegrateVelocitiesWCSPH->addConstant(accelerationLimit2);
    programIntegrateVelocitiesWCSPH->addConstant(vortexOrigin);
    programIntegrateVelocitiesWCSPH->addConstant(vortexAxis);
    programIntegrateVelocitiesWCSPH->addConstant(vortexFactor);
    programIntegrateVelocitiesWCSPH->addFunction(functionLength2);
    status &= programIntegrateVelocitiesWCSPH->build();
  }


  // PCISPH || IISPH
  if (settings->getSimulationMethod() == 1 || settings->getSimulationMethod() == 2) {
    programIntegrateVelocities = new CLProgram("clsources/integration/integrate_velocities.cl", "integrateVelocities", link);
    programClearPressureForce = new CLProgram("clsources/state_managing/clear_pressure_force.cl", "clearPressureForce", link);

    pressureForce = new CLGlobalBuffer(link, "pressureForce", CLVariableDataType::Float3Type, scene.getFluidParticlesCount(), false, true);

    programIntegrateVelocities->addGlobalBuffer(pos, 0);
    programIntegrateVelocities->addGlobalBuffer(vel, 1);
    programIntegrateVelocities->addGlobalBuffer(acc, 2);
    programIntegrateVelocities->addGlobalBuffer(pressureForce, 3);
    programIntegrateVelocities->addKernelArgument(useVortex, 4);
    programIntegrateVelocities->addConstant(deltaTime);
    programIntegrateVelocities->addConstant(accelerationLimit);
    programIntegrateVelocities->addConstant(accelerationLimit2);
    programIntegrateVelocities->addConstant(vortexOrigin);
    programIntegrateVelocities->addConstant(vortexAxis);
    programIntegrateVelocities->addConstant(vortexFactor);
    programIntegrateVelocities->addFunction(functionLength2);
    status &= programIntegrateVelocities->build();

    programClearPressureForce->addGlobalBuffer(pressureForce, 0);
    status &= programClearPressureForce->build();
  }


  // PCISPH
  if (settings->getSimulationMethod() == 1) {
    programReduceMaxDensityFluctuation = new CLProgram("clsources/helpers/reduce_max_float.cl", "reduceMaxFloat", link);
    programComputePressureForce = new CLProgram("clsources/forces/compute_pressure_force.cl", "computePressureForce", link);
    programUpdatePressures = new CLProgram("clsources/forces/update_pressures.cl", "updatePressures", link);
    programIntegrate = new CLProgram("clsources/integration/integrate.cl", "integrate", link);
    programSaveState = new CLProgram("clsources/state_managing/save_state.cl", "saveState", link);
    programLoadState = new CLProgram("clsources/state_managing/load_state.cl", "loadState", link);

    programReduceMaxDensityFluctuation->addGlobalBuffer(densityFluctuation, 0);
    programReduceMaxDensityFluctuation->addLocalBuffer(reduceScratchFloat, 1);
    programReduceMaxDensityFluctuation->addKernelArgument(reduceLength, 2);
    programReduceMaxDensityFluctuation->addGlobalBuffer(reduceResultFloat, 3);
    status &= programReduceMaxDensityFluctuation->build();

    programComputePressureForce->addGlobalBuffer(pos, 0);
    programComputePressureForce->addGlobalBuffer(density, 1);
    programComputePressureForce->addGlobalBuffer(pressure, 2);
    programComputePressureForce->addGlobalBuffer(pressureForce, 3);
    programComputePressureForce->addGlobalBuffer(neighboursFluidFluid, 4);
    programComputePressureForce->addKernelArgument(neighboursFluidFluidCountMax, 5);
    programComputePressureForce->addGlobalBuffer(solidPos, 6);
    programComputePressureForce->addGlobalBuffer(solidVolume, 7);
    programComputePressureForce->addGlobalBuffer(neighboursFluidSolid, 8);
    programComputePressureForce->addKernelArgument(neighboursFluidSolidCountMax, 9);
    programComputePressureForce->addConstant(invRadius);
    programComputePressureForce->addConstant(mass0);
    programComputePressureForce->addConstant(dWPressure);
    programComputePressureForce->addFunction(functionLength2);
    status &= programComputePressureForce->build();

    programUpdatePressures->addGlobalBuffer(pressure, 0);
    programUpdatePressures->addGlobalBuffer(density, 1);
    programUpdatePressures->addConstant(delta);
    programUpdatePressures->addConstant(deltaTime2);
    programUpdatePressures->addConstant(density0);
    status &= programUpdatePressures->build();

    programIntegrate->addGlobalBuffer(pos, 0);
    programIntegrate->addGlobalBuffer(vel, 1);
    programIntegrate->addGlobalBuffer(acc, 2);
    programIntegrate->addGlobalBuffer(pressureForce, 3);
    programIntegrate->addKernelArgument(useVortex, 4);
    programIntegrate->addConstant(deltaTime);
    programIntegrate->addConstant(accelerationLimit);
    programIntegrate->addConstant(accelerationLimit2);
    programIntegrate->addConstant(vortexOrigin);
    programIntegrate->addConstant(vortexAxis);
    programIntegrate->addConstant(vortexFactor);
    programIntegrate->addFunction(functionLength2);
    status &= programIntegrate->build();

    programSaveState->addGlobalBuffer(pos, 0);
    programSaveState->addGlobalBuffer(posTmp, 1);
    programSaveState->addGlobalBuffer(vel, 2);
    programSaveState->addGlobalBuffer(velTmp, 3);
    programSaveState->addGlobalBuffer(acc, 4);
    programSaveState->addGlobalBuffer(accTmp, 5);
    status &= programSaveState->build();

    programLoadState->addGlobalBuffer(pos, 0);
    programLoadState->addGlobalBuffer(posTmp, 1);
    programLoadState->addGlobalBuffer(vel, 2);
    programLoadState->addGlobalBuffer(velTmp, 3);
    programLoadState->addGlobalBuffer(acc, 4);
    programLoadState->addGlobalBuffer(accTmp, 5);
    status &= programLoadState->build();
  }


  // IISPH
  if (settings->getSimulationMethod() == 2) {
    programReduceSumDensityFluctuation = new CLProgram("clsources/helpers/reduce_sum_float.cl", "reduceSumFloat", link);
    programPredictVelocity = new CLProgram("clsources/forces/predict_velocity.cl", "predictVelocity", link);
    programPredictDensity = new CLProgram("clsources/forces/predict_density.cl", "predictDensity", link);
    programComputeD_ijp_j = new CLProgram("clsources/forces/compute_d_ijp_j.cl", "computeD_ijp_j", link);
    programComputeSumA = new CLProgram("clsources/forces/compute_suma.cl", "computeSumA", link);
    programComputeDensityFluctuation = new CLProgram("clsources/forces/compute_density_fluctuation.cl", "computeDensityFluctuation", link);
    programComputePressureForceIisph = new CLProgram("clsources/forces/compute_pressure_force_iisph.cl", "computePressureForceIisph", link);
    programIntegrateIisph = new CLProgram("clsources/integration/integrate_iisph.cl", "integrateIisph", link);
    programClearIisphPrediction = new CLProgram("clsources/state_managing/clear_iisph_prediction.cl", "clearIisphPrediction", link);
    programClearIisphPressures = new CLProgram("clsources/state_managing/clear_iisph_pressures.cl", "clearIisphPressures", link);

    velocityPrediction = new CLGlobalBuffer(link, "velocityPrediction", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());
    densityPrediction = new CLGlobalBuffer(link, "densityPrediction", CLVariableDataType::FloatType, scene.getFluidParticlesCount());
    d_ii = new CLGlobalBuffer(link, "d_ii", CLVariableDataType::Float3Type, scene.getFluidParticlesCount());
    a_ii = new CLGlobalBuffer(link, "a_ii", CLVariableDataType::FloatType, scene.getFluidParticlesCount());
    d_ijp_j = new CLGlobalBuffer(link, "d_ijp_j", CLVariableDataType::Float3Type, scene.getFluidParticlesCount(), false, true);
    sumA = new CLGlobalBuffer(link, "sumA", CLVariableDataType::FloatType, scene.getFluidParticlesCount());

    programReduceSumDensityFluctuation->addGlobalBuffer(densityFluctuation, 0);
    programReduceSumDensityFluctuation->addLocalBuffer(reduceScratchFloat, 1);
    programReduceSumDensityFluctuation->addKernelArgument(reduceLength, 2);
    programReduceSumDensityFluctuation->addGlobalBuffer(reduceResultFloat, 3);
    status &= programReduceSumDensityFluctuation->build();

    programPredictVelocity->addGlobalBuffer(pos, 0);
    programPredictVelocity->addGlobalBuffer(vel, 1);
    programPredictVelocity->addGlobalBuffer(acc, 2);
    programPredictVelocity->addGlobalBuffer(density, 3);
    programPredictVelocity->addGlobalBuffer(pressure, 4);
    programPredictVelocity->addGlobalBuffer(velocityPrediction, 5);
    programPredictVelocity->addGlobalBuffer(d_ii, 6);
    programPredictVelocity->addGlobalBuffer(neighboursFluidFluid, 7);
    programPredictVelocity->addKernelArgument(neighboursFluidFluidCountMax, 8);
    programPredictVelocity->addGlobalBuffer(solidPos, 9);
    programPredictVelocity->addGlobalBuffer(solidVolume, 10);
    programPredictVelocity->addGlobalBuffer(neighboursFluidSolid, 11);
    programPredictVelocity->addKernelArgument(neighboursFluidSolidCountMax, 12);
    programPredictVelocity->addConstant(deltaTime);
    programPredictVelocity->addConstant(deltaTime2);
    programPredictVelocity->addConstant(invRadius);
    programPredictVelocity->addConstant(mass0);
    programPredictVelocity->addConstant(dWPressure);
    programPredictVelocity->addFunction(functionLength2);
    status &= programPredictVelocity->build();

    programPredictDensity->addGlobalBuffer(pos, 0);
    programPredictDensity->addGlobalBuffer(density, 1);
    programPredictDensity->addGlobalBuffer(densityPrediction, 2);
    programPredictDensity->addGlobalBuffer(velocityPrediction, 3);
    programPredictDensity->addGlobalBuffer(d_ii, 4);
    programPredictDensity->addGlobalBuffer(a_ii, 5);
    programPredictDensity->addGlobalBuffer(neighboursFluidFluid, 6);
    programPredictDensity->addKernelArgument(neighboursFluidFluidCountMax, 7);
    programPredictDensity->addGlobalBuffer(solidPos, 8);
    programPredictDensity->addGlobalBuffer(solidVolume, 9);
    programPredictDensity->addGlobalBuffer(neighboursFluidSolid, 10);
    programPredictDensity->addKernelArgument(neighboursFluidSolidCountMax, 11);
    programPredictDensity->addConstant(deltaTime);
    programPredictDensity->addConstant(deltaTime2);
    programPredictDensity->addConstant(invRadius);
    programPredictDensity->addConstant(mass0);
    programPredictDensity->addConstant(dWPressure);
    programPredictDensity->addFunction(functionLength2);
    status &= programPredictDensity->build();

    programComputeD_ijp_j->addGlobalBuffer(pos, 0);
    programComputeD_ijp_j->addGlobalBuffer(d_ijp_j, 1);
    programComputeD_ijp_j->addGlobalBuffer(density, 2);
    programComputeD_ijp_j->addGlobalBuffer(pressure, 3);
    programComputeD_ijp_j->addGlobalBuffer(neighboursFluidFluid, 4);
    programComputeD_ijp_j->addKernelArgument(neighboursFluidFluidCountMax, 5);
    programComputeD_ijp_j->addConstant(deltaTime2);
    programComputeD_ijp_j->addConstant(invRadius);
    programComputeD_ijp_j->addConstant(mass0);
    programComputeD_ijp_j->addConstant(dWPressure);
    programComputeD_ijp_j->addFunction(functionLength2);
    status &= programComputeD_ijp_j->build();

    programComputeSumA->addGlobalBuffer(pos, 0);
    programComputeSumA->addGlobalBuffer(d_ii, 1);
    programComputeSumA->addGlobalBuffer(a_ii, 2);
    programComputeSumA->addGlobalBuffer(d_ijp_j, 3);
    programComputeSumA->addGlobalBuffer(sumA, 4);
    programComputeSumA->addGlobalBuffer(density, 5);
    programComputeSumA->addGlobalBuffer(pressure, 6);
    programComputeSumA->addGlobalBuffer(neighboursFluidFluid, 7);
    programComputeSumA->addKernelArgument(neighboursFluidFluidCountMax, 8);
    programComputeSumA->addGlobalBuffer(solidPos, 9);
    programComputeSumA->addGlobalBuffer(solidVolume, 10);
    programComputeSumA->addGlobalBuffer(neighboursFluidSolid, 11);
    programComputeSumA->addKernelArgument(neighboursFluidSolidCountMax, 12);
    programComputeSumA->addConstant(deltaTime2);
    programComputeSumA->addConstant(invRadius);
    programComputeSumA->addConstant(mass0);
    programComputeSumA->addConstant(dWPressure);
    programComputeSumA->addFunction(functionLength2);
    status &= programComputeSumA->build();

    programComputeDensityFluctuation->addGlobalBuffer(pressure, 0);
    programComputeDensityFluctuation->addGlobalBuffer(densityPrediction, 1);
    programComputeDensityFluctuation->addGlobalBuffer(a_ii, 2);
    programComputeDensityFluctuation->addGlobalBuffer(sumA, 3);
    programComputeDensityFluctuation->addGlobalBuffer(densityFluctuation, 4);
    programComputeDensityFluctuation->addConstant(density0);
    programComputeDensityFluctuation->addConstant(omega);
    status &= programComputeDensityFluctuation->build();

    programComputePressureForceIisph->addGlobalBuffer(pos, 0);
    programComputePressureForceIisph->addGlobalBuffer(density, 1);
    programComputePressureForceIisph->addGlobalBuffer(pressure, 2);
    programComputePressureForceIisph->addGlobalBuffer(pressureForce, 3);
    programComputePressureForceIisph->addGlobalBuffer(neighboursFluidFluid, 4);
    programComputePressureForceIisph->addKernelArgument(neighboursFluidFluidCountMax, 5);
    programComputePressureForceIisph->addGlobalBuffer(solidPos, 6);
    programComputePressureForceIisph->addGlobalBuffer(solidVolume, 7);
    programComputePressureForceIisph->addGlobalBuffer(neighboursFluidSolid, 8);
    programComputePressureForceIisph->addKernelArgument(neighboursFluidSolidCountMax, 9);
    programComputePressureForceIisph->addConstant(invRadius);
    programComputePressureForceIisph->addConstant(mass0);
    programComputePressureForceIisph->addConstant(dWPressure);
    programComputePressureForceIisph->addFunction(functionLength2);
    status &= programComputePressureForceIisph->build();

    programIntegrateIisph->addGlobalBuffer(pos, 0);
    programIntegrateIisph->addGlobalBuffer(vel, 1);
    programIntegrateIisph->addGlobalBuffer(acc, 2);
    programIntegrateIisph->addGlobalBuffer(pressureForce, 3);
    programIntegrateIisph->addGlobalBuffer(velocityPrediction, 4);
    programIntegrateIisph->addKernelArgument(useVortex, 5);
    programIntegrateIisph->addConstant(deltaTime);
    programIntegrateIisph->addConstant(accelerationLimit);
    programIntegrateIisph->addConstant(accelerationLimit2);
    programIntegrateIisph->addConstant(vortexOrigin);
    programIntegrateIisph->addConstant(vortexAxis);
    programIntegrateIisph->addConstant(vortexFactor);
    programIntegrateIisph->addFunction(functionLength2);
    status &= programIntegrateIisph->build();

    programClearIisphPrediction->addGlobalBuffer(density, 0);
    programClearIisphPrediction->addGlobalBuffer(d_ii, 1);
    programClearIisphPrediction->addGlobalBuffer(a_ii, 2);
    programClearIisphPrediction->addGlobalBuffer(densityPrediction, 3);
    status &= programClearIisphPrediction->build();

    programClearIisphPressures->addGlobalBuffer(d_ijp_j, 0);
    programClearIisphPressures->addGlobalBuffer(sumA, 1);
    status &= programClearIisphPressures->build();
  }

  settings->setCurrentDevice(link == linkCPU ? 0 : 1);
  settings->setCurrentDeviceLocalSize(link->getDevice(0)->getMaxWorkItemSize(0));

  return status;
}

void SPHsim::deleteOpenCLObjects() {
  Settings* settings = &Settings::getInstance();

  delete reduceScratchInt;
  delete reduceScratchFloat;
  delete mergeSortAux;

  delete reduceResultInt;
  delete reduceResultFloat;

  delete density;
  delete densityFluctuation;
  delete isoDensity;
  delete pressure;

  delete pos;
  delete posTmp;
  delete vel;
  delete velTmp;
  delete velXSPH;
  delete acc;
  delete accTmp;
  delete vorticity;

  delete gridCellsFluidA;
  delete gridCellsFluidB;
  delete gridCellIndicesFluid;
  delete mergeSortMergeCellsFluidLastCell;

  delete gridCellsSolidA;
  delete gridCellsSolidB;
  delete gridCellIndicesSolid;
  delete mergeSortMergeCellsSolidLastCell;

  delete neighboursFluidFluid;
  delete neighboursFluidFluidCount;
  delete neighboursFluidSolid;
  delete neighboursFluidSolidCount;
  delete neighboursSolidSolid;
  delete neighboursSolidSolidCount;

  delete solidPos;
  delete solidPosTmp;

  delete solidVolume;

  delete reduceLength;

  delete dummyElements;
  delete correctElements;

  delete mergeSortMergeMergeSize;
  delete mergeSortMergeTotalSize;

  delete usePressure;
  delete useViscosity;
  delete useVorticity;
  delete useVortex;
  delete computePressure;
  delete clearPressure;
  delete neighboursFluidFluidCountMax;
  delete neighboursFluidSolidCountMax;
  delete neighboursSolidSolidCountMax;

  delete fluidParticlesCount;
  delete solidParticlesCount;
  delete deltaTime;
  delete deltaTime2;

  delete delta;
  delete omega;
  delete gravity;
  delete radius;
  delete radius2;
  delete invRadius;
  delete invRadius2;
  delete accelerationLimit;
  delete accelerationLimit2;
  delete wDensity;
  delete dWDensity;
  delete ddWViscosity;
  delete dWPressure;
  delete density0;
  delete mass0;
  delete viscosity0;
  delete solidViscosity0;
  delete stiffness;
  delete negativePressureFactor;
  delete seventh;
  delete surfaceTension;
  delete boundaryStiffness;
  delete vorticityFactor;
  delete vortexOrigin;
  delete vortexAxis;
  delete vortexFactor;

  delete functionDistance2;
  delete functionLength2;
  delete functionZOrder;

  delete programReduceMaxFluidFluidCount;
  delete programReduceMaxFluidSolidCount;
  delete programReduceMaxSolidSolidCount;
  delete programComputeNeighboursFluidCount;
  delete programComputeNeighboursSolidCount;
  delete programCreateCellsFluid;
  delete programCreateCellsSolid;
  delete programCorrectCellsFluid;
  delete programCorrectCellsSolid;
  delete programMergeSortCellsFluid;
  delete programMergeSortCellsSolid;
  delete programMergeSortMergeCellsFluidA;
  delete programMergeSortMergeCellsFluidB;
  delete programMergeSortMergeCellsSolidA;
  delete programMergeSortMergeCellsSolidB;
  delete programSwapFluid;
  delete programSwapSolid;
  delete programClearCellIndicesFluid;
  delete programClearCellIndicesSolid;
  delete programCreateCellIndicesFluid;
  delete programCreateCellIndicesSolid;
  delete programFindNeighboursFluid;
  delete programFindNeighboursSolid;
  delete programFindNeighboursFluidN2;
  delete programFindNeighboursSolidN2;
  delete programComputeSolidVolume;
  delete programComputeDensity;
  delete programNormalizeDensityFirstPass;
  delete programNormalizeDensitySecondPass;
  delete programComputeForce;
  delete programIntegratePositions;
  delete programApplyXSPHFirstPass;
  delete programApplyXSPHSecondPass;
  delete programClearState;


  // WCSPH
  if (settings->getSimulationMethod() == 0) {
    delete programIntegrateVelocitiesWCSPH;
  }


  // PCISPH || IISPH
  if (settings->getSimulationMethod() == 1 || settings->getSimulationMethod() == 2) {
    delete pressureForce;

    delete programIntegrateVelocities;
    delete programClearPressureForce;
  }


  // PCISPH
  if (settings->getSimulationMethod() == 1) {
    delete programReduceMaxDensityFluctuation;
    delete programComputePressureForce;
    delete programUpdatePressures;
    delete programIntegrate;
    delete programSaveState;
    delete programLoadState;
  }


  // IISPH
  if (settings->getSimulationMethod() == 2) {
    delete velocityPrediction;
    delete densityPrediction;
    delete d_ii;
    delete a_ii;
    delete d_ijp_j;
    delete sumA;

    delete programReduceSumDensityFluctuation;
    delete programPredictVelocity;
    delete programPredictDensity;
    delete programComputeD_ijp_j;
    delete programComputeSumA;
    delete programComputeDensityFluctuation;
    delete programComputePressureForceIisph;
    delete programIntegrateIisph;
    delete programClearIisphPrediction;
    delete programClearIisphPressures;
  }
}

bool SPHsim::setSceneDamBreak() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(14.5, 14.5, 1), Vector<3, float>(16, 16, 16), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.09f, 0.09f, 0.09f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  return scene.initScene("Scene Dam Break", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setSceneDamBreakObstacle() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(18.1, 18.1, 1), Vector<3, float>(16, 16, 16), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.1f, 0.1f, 0.1f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  Object obstacle;
  if (!obstacle.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  obstacle.scale(Vector<3, float>(0.05f, 0.05f, 0.1f));
  obstacle.translate(Vector<3, float>(0.05f, 0.05f, 0.1f));

  std::vector<Vector<3, float>> obstacleParticles = obstacle.generateSurfaceParticles(restDistance / 2.0f);
  solidParticles.insert(solidParticles.end(), obstacleParticles.begin(), obstacleParticles.end());

  objects.push_back(obstacle);

  return scene.initScene("Scene Dam Break Obstacle", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setSceneDamBreakTwoFluids() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(18, 18, 1), Vector<3, float>(16, 16, 16), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  std::vector<Vector<3, float>> another = SPHsim::createFluidBox(Vector<3, float>(1, 1, 1), Vector<3, float>(16, 16, 16), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);
  fluidParticles.insert(fluidParticles.end(), another.begin(), another.end());

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.1f, 0.1f, 0.1f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  return scene.initScene("Scene Dam Break Two Fluids", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setSceneOcean() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(1, 39.5, 1), Vector<3, float>(32.5, 20, 20), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.1f, 0.175f, 0.1f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  Object obstacle;
  if (!obstacle.loadFromFile("meshes//cylinder.obj")) {
    return false;
  }
  obstacle.scale(Vector<3, float>(0.05f, 0.05f, 0.05f));
  obstacle.translate(Vector<3, float>(0.1f, 0.125f, 0.0f));

  std::vector<Vector<3, float>> obstacleParticles = obstacle.generateSurfaceParticles(restDistance / 2.0f);
  solidParticles.insert(solidParticles.end(), obstacleParticles.begin(), obstacleParticles.end());

  objects.push_back(obstacle);

  return scene.initScene("Scene Ocean", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setSceneOceanBig() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(1, 86.4, 1), Vector<3, float>(66.9, 50, 50), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.2f, 0.4f, 0.2f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  std::vector<Vector<3, float>> obstacleParticles;

  Object obstacle1;
  if (!obstacle1.loadFromFile("meshes//cylinder.obj")) {
    return false;
  }
  obstacle1.scale(Vector<3, float>(0.075f, 0.075f, 0.1f));
  obstacle1.translate(Vector<3, float>(0.1f, 0.15f, 0.0f));

  obstacleParticles = obstacle1.generateSurfaceParticles(restDistance / 2.0f);
  solidParticles.insert(solidParticles.end(), obstacleParticles.begin(), obstacleParticles.end());

  objects.push_back(obstacle1);

  Object obstacle2;
  if (!obstacle2.loadFromFile("meshes//cylinder.obj")) {
    return false;
  }
  obstacle2.scale(Vector<3, float>(0.05f, 0.05f, 0.125f));
  obstacle2.translate(Vector<3, float>(0.32f, 0.25f, 0.0f));

  obstacleParticles = obstacle2.generateSurfaceParticles(restDistance / 2.0f);
  solidParticles.insert(solidParticles.end(), obstacleParticles.begin(), obstacleParticles.end());

  objects.push_back(obstacle2);

  Object obstacle3;
  if (!obstacle3.loadFromFile("meshes//cylinder.obj")) {
    return false;
  }
  obstacle3.scale(Vector<3, float>(0.04f, 0.04f, 0.15f));
  obstacle3.translate(Vector<3, float>(0.2f, 0.375f, 0.0f));

  obstacleParticles = obstacle3.generateSurfaceParticles(restDistance / 2.0f);
  solidParticles.insert(solidParticles.end(), obstacleParticles.begin(), obstacleParticles.end());

  objects.push_back(obstacle3);

  return scene.initScene("Scene Ocean Big", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setSceneDrop() {
  fluidParticles.clear();
  fluidParticles = SPHsim::createFluidBox(Vector<3, float>(8, 8, 16), Vector<3, float>(16, 16, 16), Vector<3, float>(0, 0, 0), distance0Host / 10, 10000);

  Object aquarium;
  std::vector<Object> objects;

  if (!aquarium.loadFromFile("meshes//cube.obj")) {
    return false;
  }
  Vector<3, float> aquariumScale(0.09f, 0.09f, 0.11f);
  aquarium.scale(aquariumScale);
  aquarium.translate(aquariumScale);

  vortexOriginHost = Vector<3, float>(0.5f * aquariumScale.x, 0.5f * aquariumScale.y, 0);

  solidParticles.clear();
  solidParticles = aquarium.generateSurfaceParticles(restDistance / 2.0f);

  return scene.initScene("Scene Drop", aquarium, objects, fluidParticles.size(), solidParticles.size());
}

bool SPHsim::setCurrentScene() {
  Settings* settings = &Settings::getInstance();

  return (this->*scenesPointers[settings->getCurrentScene()])();
}

void SPHsim::initScene() {
  Settings* settings = &Settings::getInstance();

  simTimeHost = 0.0f;
  settings->setSimulationTime(simTimeHost);
  framesHost = 0;

  initializeParticles(fluidParticles, solidParticles);
  gridCellIndicesFluid->reallocate(1);
  gridCellIndicesSolid->reallocate(1);
  computeGridFluid();
  computeGridSolid();
  reallocateNeighboursBuffers();

  programFindNeighboursSolid->enqueue();
  programFindNeighboursSolid->finish();

  programComputeSolidVolume->enqueue();
  programComputeSolidVolume->finish();
}

bool SPHsim::createScene() {
  Settings* settings = &Settings::getInstance();
  Log* log = Log::getInstance();
  
  log->writeLine(SPHDebug, "Creating SPH simulation scene...");

  if (settings->getCurrentScene() < 0 || settings->getCurrentScene() > scenesPointers.size()) {
    settings->setCurrentScene(0);
  }

  minIterationsPCISPH = 4;
  minIterationsIISPH = 2;
  maxIterationsIISPH = 20;
  deltaTimeHost = 0.001f;

  omegaHost = 0.5f;
  gravityHost = -9.81f;
  radiusHost = 0.01f;
  accelerationLimitHost = 400.0f;
  wDensityHost = 315.0f / (64.0f * Consts::PI * pow(radiusHost, 3));
  dWDensityHost = -945.0f / (32.0f * Consts::PI * pow(radiusHost, 5));
  ddWViscosityHost = 45.0f / (Consts::PI * pow(radiusHost, 5));
  dWPressureHost = -45.0f / (Consts::PI * pow(radiusHost, 4));
  density0Host = 1000.0f;
  viscosity0Host = 0.2f;
  solidViscosity0Host = 0.01f;
  mass0Host = 0.0002f;
  volume0Host = mass0Host / density0Host;
  distance0Host = powf(volume0Host, 1.0f / 3.0f);
  stiffnessHost = 40.0f;
  negativePressureFactorHost = 0.002f;
  surfaceTensionHost = 0.00007f;
  boundaryStiffnessHost = 40000.0f;
  vorticityFactorHost = 1.1f;

  restDistance = 0.01f;

  precomputeDelta();

  bool sceneSet = setCurrentScene();

  if (sceneSet) {
    vortexAxisHost = Vector<3, float>(0, 0, 1);
    vortexFactorHost = 0.1f;

    clsys = CLSystem::getInstance();

    CLPlatform* platform = clsys->getPlatform(0);

    if (platform == NULL) {
      log->writeLine(SPHError, "No OpenCL platform found!");
      return false;
    }

    std::vector<CLDevice*> devices = platform->getDevices();
    CLDevice* gpuDevice = NULL, *cpuDevice = NULL;

    for (int i = 0; i < devices.size(); i++) {
      if (devices[i]->isCPU()) {
        cpuDevice = devices[i];
      }
      else if (gpuDevice == NULL && devices[i]->isGPU()) {
        gpuDevice = devices[i];
      }
    }

    if (cpuDevice == NULL) {
      log->writeLine(SPHError, "No OpenCL CPU device found!");
      return false;
    }

    if (gpuDevice == NULL) {
      log->writeLine(SPHError, "No OpenCL GPU device found!");
      return false;
    }

    linkCPU = new CLLink(cpuDevice);
    linkGPU = new CLLink(gpuDevice);

    allCLProgramsBuilt = createOpenCLObjects((settings->getCurrentDevice() == 0 ? linkCPU : linkGPU));

    if (allCLProgramsBuilt) {
      initScene();
    }
  }

  return sceneSet;
}

void SPHsim::resetScene() {
  Settings* settings = &Settings::getInstance();
  
  if (allCLProgramsBuilt) {
    if (setCurrentScene()) {
      initScene();
    }
  }
}

void SPHsim::deleteScene() {
  Log* log = Log::getInstance();

  log->writeLine(SPHDebug, "Deleting SPH simulation scene...");

  deleteOpenCLObjects();

  delete linkCPU;
  delete linkGPU;

  delete clsys;

  scene.freeScene();
}

bool SPHsim::nextScene() {
  Settings* settings = &Settings::getInstance();
  Log* log = Log::getInstance();

  log->writeLine(SPHDebug, "Switching next SPH simulation scene...");

  settings->setCurrentScene((settings->getCurrentScene() + 1) % scenesPointers.size());

  deleteOpenCLObjects();
  bool sceneSet = setCurrentScene();
  allCLProgramsBuilt = createOpenCLObjects((settings->getCurrentDevice() == 0 ? linkCPU : linkGPU));

  if (allCLProgramsBuilt) {
    if (sceneSet) {
      initScene();
    }
  }
  
  return sceneSet;
}

bool SPHsim::previousScene() {
  Settings* settings = &Settings::getInstance();
  Log* log = Log::getInstance();

  log->writeLine(SPHDebug, "Switching previous SPH simulation scene...");

  settings->setCurrentScene((settings->getCurrentScene() > 0) ? settings->getCurrentScene() - 1 : scenesPointers.size() - 1);

  deleteOpenCLObjects();
  bool sceneSet = setCurrentScene();
  allCLProgramsBuilt = createOpenCLObjects((settings->getCurrentDevice() == 0 ? linkCPU : linkGPU));
    
  if (allCLProgramsBuilt) {
    if (sceneSet) {
      initScene();
    }
  }

  return sceneSet;
}

void SPHsim::switchDevice() {
  Settings* settings = &Settings::getInstance();
  Log* log = Log::getInstance();

  log->writeLine(SPHDebug, "Switching OpenCL device to " + ((settings->getCurrentDevice() == 0) ? std::string("GPU") : std::string("CPU")));

  if ((settings->getCurrentDevice() == 0 && linkGPU != NULL) || (settings->getCurrentDevice() == 1 && linkCPU != NULL)) {
    deleteOpenCLObjects();
    allCLProgramsBuilt = createOpenCLObjects((settings->getCurrentDevice() == 0 ? linkGPU : linkCPU));

    resetScene();
  }
}

void SPHsim::switchSimulationMethod() {
  Settings* settings = &Settings::getInstance();
  Log* log = Log::getInstance();

  log->writeLine(SPHDebug, "Switching SPH simulation method...");

  deleteOpenCLObjects();
  settings->setSimulationMethod((settings->getSimulationMethod() + 1) % 3);
  allCLProgramsBuilt = createOpenCLObjects((settings->getCurrentDevice() == 0 ? linkCPU : linkGPU));

  resetScene();
}

void SPHsim::doStep() {
  Settings* settings = &Settings::getInstance();

  if (allCLProgramsBuilt) {
    double t, overall;

    settings->setMergeSort1Time(0);
    settings->setMergeSort2Time(0);
    settings->setComputeGridTime(0);
    settings->setNeighboursTime(0);
    settings->setComputeDensityTime(0);
    settings->setComputeForceTime(0);
    settings->setIntegrateTime(0);
    settings->setStateManagingTime(0);

    framesHost++;
    simTimeHost += deltaTimeHost;
    settings->setSimulationTime(simTimeHost);

    overall = clock();

    if (settings->getComputeGridEveryFrame() || framesHost % 10 == 0) {
      t = clock();
      computeGridFluid();
      settings->setComputeGridTime(clock() - t);
    }

    t = clock();
    programFindNeighboursFluid->enqueue();
    programFindNeighboursFluid->finish();
    settings->setNeighboursTime(clock() - t);

    //////////////////
    // WCSPH
    //////////////////
    if (settings->getSimulationMethod() == 0) {
      t = clock();
      clearPressure->setScalar(1);
      programClearState->enqueue();
      programClearState->finish();
      settings->setStateManagingTime(clock() - t);

      t = clock();
      useVorticity->setScalar(1);
      programComputeDensity->enqueue();
      programComputeDensity->finish();

      programNormalizeDensityFirstPass->enqueue();
      programNormalizeDensityFirstPass->finish();

      computePressure->setScalar(1);
      programNormalizeDensitySecondPass->enqueue();
      programNormalizeDensitySecondPass->finish();
      settings->setComputeDensityTime(clock() - t);

      t = clock();
      usePressure->setScalar(1);
      useViscosity->setScalar(1);
      useVorticity->setScalar(1);
      programComputeForce->enqueue();
      programComputeForce->finish();
      settings->setComputeForceTime(clock() - t);

      t = clock();
      useVortex->setScalar(settings->getVortex());
      programIntegrateVelocitiesWCSPH->enqueue();
      programIntegrateVelocitiesWCSPH->finish();

      programApplyXSPHFirstPass->enqueue();
      programApplyXSPHFirstPass->finish();

      programApplyXSPHSecondPass->enqueue();
      programApplyXSPHSecondPass->finish();

      programIntegratePositions->enqueue();
      programIntegratePositions->finish();
      settings->setIntegrateTime(clock() - t);
    }
    //////////////////
    // PCISPH
    //////////////////
    else if (settings->getSimulationMethod() == 1) {
      t = clock();
      clearPressure->setScalar(1);
      programClearState->enqueue();
      programClearState->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      t = clock();
      useVorticity->setScalar(1);
      programComputeDensity->enqueue();
      programComputeDensity->finish();

      programNormalizeDensityFirstPass->enqueue();
      programNormalizeDensityFirstPass->finish();

      computePressure->setScalar(0);
      programNormalizeDensitySecondPass->enqueue();
      programNormalizeDensitySecondPass->finish();

      reduceLength->setScalar(scene.getFluidParticlesCount());
      programReduceMaxDensityFluctuation->enqueue(reduceLocalSize, reduceLocalSize);
      programReduceMaxDensityFluctuation->finish();

      reduceResultFloat->download(true, true);
      maxDensityFluctuation = (float)reduceResultFloat->getScalar(0);
      settings->setComputeDensityTime(clock() - t);

      t = clock();
      usePressure->setScalar(0);
      useViscosity->setScalar(1);
      useVorticity->setScalar(1);
      programComputeForce->enqueue();
      programComputeForce->finish();
      settings->setComputeForceTime(settings->getComputeForceTime() + clock() - t);

      t = clock();
      programClearPressureForce->enqueue();
      programClearPressureForce->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      t = clock();
      programSaveState->enqueue();
      programSaveState->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      int iter = 0;
      while ((iter < minIterationsPCISPH || maxDensityFluctuation > 10) && iter < 10) {
        t = clock();
        clearPressure->setScalar(0);
        programClearState->enqueue();
        programClearState->finish();
        settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

        t = clock();
        programLoadState->enqueue();
        programLoadState->finish();
        settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

        t = clock();
        useVortex->setScalar(settings->getVortex());
        programIntegrate->enqueue();
        programIntegrate->finish();
        settings->setIntegrateTime(settings->getIntegrateTime() + clock() - t);

        t = clock();
        programComputeDensity->enqueue();
        programComputeDensity->finish();

        programNormalizeDensityFirstPass->enqueue();
        programNormalizeDensityFirstPass->finish();

        computePressure->setScalar(0);
        programNormalizeDensitySecondPass->enqueue();
        programNormalizeDensitySecondPass->finish();

        reduceLength->setScalar(scene.getFluidParticlesCount());
        programReduceMaxDensityFluctuation->enqueue(reduceLocalSize, reduceLocalSize);
        programReduceMaxDensityFluctuation->finish();

        reduceResultFloat->download(true, true);
        maxDensityFluctuation = (float)reduceResultFloat->getScalar(0);

        programUpdatePressures->enqueue();
        programUpdatePressures->finish();
        settings->setComputeDensityTime(settings->getComputeDensityTime() + clock() - t);

        t = clock();
        programClearPressureForce->enqueue();
        programClearPressureForce->finish();
        settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

        t = clock();
        programComputePressureForce->enqueue();
        programComputePressureForce->finish();
        settings->setComputeForceTime(settings->getComputeForceTime() + clock() - t);

        iter++;
      }

      t = clock();
      programLoadState->enqueue();
      programLoadState->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      t = clock();
      useVortex->setScalar(settings->getVortex());
      programIntegrateVelocities->enqueue();
      programIntegrateVelocities->finish();

      programApplyXSPHFirstPass->enqueue();
      programApplyXSPHFirstPass->finish();

      programApplyXSPHSecondPass->enqueue();
      programApplyXSPHSecondPass->finish();

      programIntegratePositions->enqueue();
      programIntegratePositions->finish();
      settings->setIntegrateTime(clock() - t);
    }
    //////////////////
    // IISPH
    //////////////////
    else if (settings->getSimulationMethod() == 2) {
      t = clock();
      clearPressure->setScalar(0);
      programClearState->enqueue();
      programClearState->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      t = clock();
      useVorticity->setScalar(0);
      programComputeDensity->enqueue();
      programComputeDensity->finish();

      programNormalizeDensityFirstPass->enqueue();
      programNormalizeDensityFirstPass->finish();

      computePressure->setScalar(0);
      programNormalizeDensitySecondPass->enqueue();
      programNormalizeDensitySecondPass->finish();

      reduceLength->setScalar(scene.getFluidParticlesCount());
      programReduceSumDensityFluctuation->enqueue(reduceLocalSize, reduceLocalSize);
      programReduceSumDensityFluctuation->finish();

      reduceResultFloat->download(true, true);
      avgDensityFluctuation = (float)reduceResultFloat->getScalar(0) / (float)scene.getFluidParticlesCount();
      settings->setComputeDensityTime(clock() - t);

      t = clock();
      usePressure->setScalar(0);
      useViscosity->setScalar(0);
      useVorticity->setScalar(0);
      programComputeForce->enqueue();
      programComputeForce->finish();
      settings->setComputeForceTime(clock() - t);

      t = clock();
      programClearIisphPrediction->enqueue();
      programClearIisphPrediction->finish();

      programPredictVelocity->enqueue();
      programPredictVelocity->finish();

      programPredictDensity->enqueue();
      programPredictDensity->finish();
      settings->setComputeDensityTime(clock() - t);

      int iter = 0;
      while ((iter < minIterationsIISPH || avgDensityFluctuation > 5) && iter < maxIterationsIISPH) {
        iter++;

        t = clock();
        programClearIisphPressures->enqueue();
        programClearIisphPressures->finish();
        settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

        t = clock();
        programComputeD_ijp_j->enqueue();
        programComputeD_ijp_j->finish();

        programComputeSumA->enqueue();
        programComputeSumA->finish();

        programComputeDensityFluctuation->enqueue();
        programComputeDensityFluctuation->finish();

        reduceLength->setScalar(scene.getFluidParticlesCount());
        programReduceSumDensityFluctuation->enqueue(reduceLocalSize, reduceLocalSize);
        programReduceSumDensityFluctuation->finish();
        settings->setComputeForceTime(settings->getComputeForceTime() + clock() - t);

        reduceResultFloat->download(true, true);
        avgDensityFluctuation = (float)reduceResultFloat->getScalar(0) / (float)scene.getFluidParticlesCount();
      }

      t = clock();
      programClearPressureForce->enqueue();
      programClearPressureForce->finish();
      settings->setStateManagingTime(settings->getStateManagingTime() + clock() - t);

      t = clock();
      programComputePressureForceIisph->enqueue();
      programComputePressureForceIisph->finish();
      
      useVortex->setScalar(settings->getVortex());
      programIntegrateIisph->enqueue();
      programIntegrateIisph->finish();

      programApplyXSPHFirstPass->enqueue();
      programApplyXSPHFirstPass->finish();

      programApplyXSPHSecondPass->enqueue();
      programApplyXSPHSecondPass->finish();

      programIntegratePositions->enqueue();
      programIntegratePositions->finish();
      settings->setIntegrateTime(clock() - t);
    }

    settings->setOverallComputationTime(clock() - overall);

    t = clock();
    pos->download(true, true);
    if (settings->getFluidParticlesVisualization() == 1 || settings->getFluidParticlesVisualization() == 2) {
      vel->download(true, true);
    }
    settings->setDownloadTime(clock() - t);

    if (framesHost % 10 == 0) {
      t = clock();
      reallocateNeighboursBuffers(true);
      settings->setReallocNeighboursBufferTime(clock() - t);
    }

    scene.setFluidParticlesData(pos->getData(), pos->getMemorySize());
    scene.setFluidParticlesDataOther(vel->getData(), vel->getMemorySize());
  }
}